# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" This module provides functions for registering interaction handlers
and location solvers for common Wx GUI components.
"""

from traitsui.testing.tester.command import KeyClick, KeySequence, MouseClick
from traitsui.testing.tester.query import DisplayedText
from traitsui.testing.tester._ui_tester_registry.wx import _interaction_helpers


def register_editable_textbox_handlers(registry, target_class, widget_getter):
    """Register common interactions for an editable textbox (in Wx)

    Parameters
    ----------
    registry : TargetRegistry
        The registry being registered to.
    target_class : subclass of type
        The type of target being wrapped in a UIWrapper on which the
        interaction will be performed.
    widget_getter : callable(wrapper: UIWrapper) -> wx.TextCtrl
        A callable to return a wx.TextCtrl
    """
    handlers = [
        (
            KeySequence,
            (
                lambda wrapper, interaction: _interaction_helpers.key_sequence_text_ctrl(
                    widget_getter(wrapper), interaction, wrapper.delay
                )
            ),
        ),
        (
            KeyClick,
            (
                lambda wrapper, interaction: _interaction_helpers.key_click_text_entry(
                    widget_getter(wrapper), interaction, wrapper.delay
                )
            ),
        ),
        (
            MouseClick,
            (
                lambda wrapper, _: _interaction_helpers.mouse_click_object(
                    control=widget_getter(wrapper), delay=wrapper.delay
                )
            ),
        ),
        (DisplayedText, lambda wrapper, _: widget_getter(wrapper).GetValue()),
    ]
    for interaction_class, handler in handlers:
        registry.register_interaction(
            target_class=target_class,
            interaction_class=interaction_class,
            handler=handler,
        )



# (C) Copyright 2008-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from pyface.qt import QtGui
from pyface.color import Color as PyfaceColor
from traits.api import HasStrictTraits, TraitError

from traitsui.qt4.color_trait import PyQtColor


class ObjectWithColor(HasStrictTraits):

    color = PyQtColor()


class ObjectWithColorAllowsNone(HasStrictTraits):

    color = PyQtColor(allow_none=True)


class TestPyQtColor(unittest.TestCase):

    def test_default(self):
        obj = ObjectWithColor()

        self.assertIsInstance(obj.color, QtGui.QColor)
        self.assertEqual(obj.color.getRgb(), (255, 255, 255, 255))
        self.assertIsInstance(obj.color_, QtGui.QColor)
        self.assertEqual(obj.color_.getRgb(), (255, 255, 255, 255))

    def test_tuple_rgb(self):
        obj = ObjectWithColor(color=(0, 128, 255))

        self.assertIsInstance(obj.color, QtGui.QColor)
        self.assertEqual(obj.color.getRgb(), (0, 128, 255, 255))
        self.assertIsInstance(obj.color_, QtGui.QColor)
        self.assertEqual(obj.color_.getRgb(), (0, 128, 255, 255))

    def test_tuple_rgba(self):
        obj = ObjectWithColor(color=(0, 128, 255, 64))

        self.assertIsInstance(obj.color, QtGui.QColor)
        self.assertEqual(obj.color.getRgb(), (0, 128, 255, 64))
        self.assertIsInstance(obj.color_, QtGui.QColor)
        self.assertEqual(obj.color_.getRgb(), (0, 128, 255, 64))

    def test_name_string(self):
        obj = ObjectWithColor(color="rebeccapurple")

        self.assertIsInstance(obj.color, QtGui.QColor)
        self.assertEqual(obj.color.getRgb(), (0x66, 0x33, 0x99, 0xff))
        self.assertIsInstance(obj.color_, QtGui.QColor)
        self.assertEqual(obj.color_.getRgb(), (0x66, 0x33, 0x99, 0xff))

    def test_name_string_with_space(self):
        obj = ObjectWithColor(color="rebecca purple")

        self.assertIsInstance(obj.color, QtGui.QColor)
        self.assertEqual(obj.color.getRgb(), (0x66, 0x33, 0x99, 0xff))
        self.assertIsInstance(obj.color_, QtGui.QColor)
        self.assertEqual(obj.color_.getRgb(), (0x66, 0x33, 0x99, 0xff))

    def test_rgb_string(self):
        obj = ObjectWithColor(color="(0, 128, 255)")

        self.assertIsInstance(obj.color, QtGui.QColor)
        self.assertEqual(obj.color.getRgb(), (0, 128, 255, 255))
        self.assertIsInstance(obj.color_, QtGui.QColor)
        self.assertEqual(obj.color_.getRgb(), (0, 128, 255, 255))

    def test_rgba_string(self):
        obj = ObjectWithColor(color="(0, 128, 255, 64)")

        self.assertIsInstance(obj.color, QtGui.QColor)
        self.assertEqual(obj.color.getRgb(), (0, 128, 255, 64))
        self.assertIsInstance(obj.color_, QtGui.QColor)
        self.assertEqual(obj.color_.getRgb(), (0, 128, 255, 64))

    def test_rgb_int_string_3(self):
        obj = ObjectWithColor(color="#08f")

        self.assertIsInstance(obj.color, QtGui.QColor)
        self.assertEqual(obj.color.getRgb(), (0x00, 0x88, 0xff, 0xff))
        self.assertIsInstance(obj.color_, QtGui.QColor)
        self.assertEqual(obj.color_.getRgb(), (0x00, 0x88, 0xff, 0xff))

    def test_rgb_int_string_6(self):
        obj = ObjectWithColor(color="#0088ff")

        self.assertIsInstance(obj.color, QtGui.QColor)
        self.assertEqual(obj.color.getRgb(), (0x00, 0x88, 0xff, 0xff))
        self.assertIsInstance(obj.color_, QtGui.QColor)
        self.assertEqual(obj.color_.getRgb(), (0x00, 0x88, 0xff, 0xff))

    def test_rgb_int_string_12(self):
        obj = ObjectWithColor(color="#00008888ffff")

        self.assertIsInstance(obj.color, QtGui.QColor)
        self.assertEqual(obj.color.getRgb(), (0x00, 0x88, 0xff, 0xff))
        self.assertIsInstance(obj.color_, QtGui.QColor)
        self.assertEqual(obj.color_.getRgb(), (0x00, 0x88, 0xff, 0xff))

    def test_rgb_int(self):
        obj = ObjectWithColor(color=0x0088ff)

        self.assertIsInstance(obj.color, QtGui.QColor)
        self.assertEqual(obj.color.getRgb(), (0x00, 0x88, 0xff, 0xff))
        self.assertIsInstance(obj.color_, QtGui.QColor)
        self.assertEqual(obj.color_.getRgb(), (0x00, 0x88, 0xff, 0xff))

    def test_qcolor(self):
        obj = ObjectWithColor(color=QtGui.QColor(0, 128, 255, 64))

        self.assertIsInstance(obj.color, QtGui.QColor)
        self.assertEqual(obj.color.getRgb(), (0, 128, 255, 64))
        self.assertIsInstance(obj.color_, QtGui.QColor)
        self.assertEqual(obj.color_.getRgb(), (0, 128, 255, 64))

    def test_pyface_color(self):
        obj = ObjectWithColor(color=PyfaceColor(rgba=(0.0, 0.5, 1.0, 0.25)))

        self.assertIsInstance(obj.color, QtGui.QColor)
        self.assertEqual(obj.color.getRgb(), (0, 128, 255, 64))
        self.assertIsInstance(obj.color_, QtGui.QColor)
        self.assertEqual(obj.color_.getRgb(), (0, 128, 255, 64))

    def test_default_none(self):
        obj = ObjectWithColorAllowsNone(color=None)

        self.assertIsNone(obj.color)
        self.assertIsNone(obj.color_)

    def test_bad_color(self):
        with self.assertRaises(TraitError):
            ObjectWithColor(color="not a color")

    def test_bad_tuple(self):
        with self.assertRaises(TraitError):
            ObjectWithColor(color=(0xff, 0xff))

    def test_bad_tuple_not_int(self):
        with self.assertRaises(TraitError):
            ObjectWithColor(color=("not an int", 0xff, 0xff))

    def test_bad_tuple_string(self):
        with self.assertRaises(TraitError):
            ObjectWithColor(color="(0xff, 0xff)")



# (C) Copyright 2009-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# ------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD license.
# However, when used with the GPL version of PyQt the additional terms
# described in the PyQt GPL exception also apply

#
# Author: Riverbank Computing Limited
# ------------------------------------------------------------------------------


from pyface.qt import QtCore, QtGui

from .editor import Editor

from pyface.api import HeadingText


class SimpleEditor(Editor):

    def init(self, parent):
        """Finishes initializing the editor by creating the underlying toolkit
        widget.
        """
        if isinstance(parent, QtGui.QLayout):
            parent = parent.parentWidget()
        self._control = HeadingText(parent=parent, create=False)
        self._control.create()
        self.control = self._control.control
        if self.factory.allow_selection:
            flags = (
                self.control.textInteractionFlags()
                | QtCore.Qt.TextInteractionFlag.TextSelectableByMouse
            )
            self.control.setTextInteractionFlags(flags)
        self.set_tooltip()

    def update_editor(self):
        """Updates the editor when the object trait changes external to the
        editor.
        """
        self._control.text = self.str_value

    def dispose(self):
        """Cleanly dispose of the editor.

        This ensures that the wrapped Pyface Widget is cleaned-up.
        """
        if self._control is not None:
            self._control.destroy()
        super().dispose()


CustomEditor = SimpleEditor
ReadonlyEditor = SimpleEditor
TextEditor = SimpleEditor



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Edit a List of Strings

Simple demonstration of the ListStrEditor, which can be used for editing and
displaying lists of strings, or other data that can be logically mapped to a
list of strings.
"""

from traits.api import HasTraits, List, Str
from traitsui.api import View, Item, ListStrEditor

# -- ShoppingListDemo Class -----------------------------------------------


class ShoppingListDemo(HasTraits):

    # The list of things to buy at the store:
    shopping_list = List(Str)

    # -- Traits View Definitions ----------------------------------------------

    view = View(
        Item(
            'shopping_list',
            show_label=False,
            editor=ListStrEditor(title='Shopping List', auto_add=True),
        ),
        title='Shopping List',
        width=0.2,
        height=0.5,
        resizable=True,
    )


# -- Set up the Demo ------------------------------------------------------

demo = ShoppingListDemo(
    shopping_list=[
        'Carrots',
        'Potatoes (5 lb. bag)',
        'Cocoa Puffs',
        'Ice Cream (French Vanilla)',
        'Peanut Butter',
        'Whole wheat bread',
        'Ground beef (2 lbs.)',
        'Paper towels',
        'Soup (3 cans)',
        'Laundry detergent',
    ]
)

# Run the demo (in invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# FIXME: PopupEditor is a proxy class defined here just for backward
# compatibility. The class (which represents the editor factory) has been moved
# to the traitsui.editors.list_editor file.
from traitsui.editors.popup_editor import (
    _PopupEditor as BasePopupEditor,
    PopupEditor,
)

from .ui_editor import UIEditor

# -------------------------------------------------------------------------
#  '_PopupEditor' class:
# -------------------------------------------------------------------------


class _PopupEditor(BasePopupEditor, UIEditor):
    pass



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" A renderer which displays a progress bar. """

# System library imports
import sys
from pyface.qt import QtCore, QtGui

# ETS imports
from traitsui.qt4.table_editor import TableDelegate


class ProgressRenderer(TableDelegate):
    """A renderer which displays a progress bar."""

    # -------------------------------------------------------------------------
    #  QAbstractItemDelegate interface
    # -------------------------------------------------------------------------

    def paint(self, painter, option, index):
        """Paint the progressbar."""
        # Get the column and object
        column = index.model()._editor.columns[index.column()]
        obj = index.data(QtCore.Qt.ItemDataRole.UserRole)

        # set up progress bar options
        progress_bar_option = QtGui.QStyleOptionProgressBar()
        progress_bar_option.rect = option.rect
        progress_bar_option.minimum = column.get_minimum(obj)
        progress_bar_option.maximum = column.get_maximum(obj)
        progress_bar_option.progress = int(column.get_raw_value(obj))
        progress_bar_option.textVisible = column.get_text_visible()
        progress_bar_option.text = column.get_value(obj)

        # Draw it
        style = QtGui.QApplication.instance().style()
        if sys.platform == "darwin":
            # Save painter state, translate painter to cell location, and then
            # restore painter state after drawing to solve enthought/traitsui#964
            # This seems to be mac only.
            # ref: https://forum.qt.io/topic/105375/qitemdelegate-for-drawing-progress-bar-working-but-won-t-move-off-origin  # noqa: E501
            painter.save()
            painter.translate(option.rect.left(), option.rect.top())
            style.drawControl(
                QtGui.QStyle.ControlElement.CE_ProgressBar, progress_bar_option, painter
            )
            painter.restore()
        else:
            # For non-mac platforms, just draw.
            style.drawControl(
                QtGui.QStyle.ControlElement.CE_ProgressBar, progress_bar_option, painter
            )



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines KeyBinding and KeyBindings classes, which manage the mapping of
    keystroke events into method calls on controller objects that are supplied
    by the application.
"""

from traits.api import (
    Any,
    Event,
    HasPrivateTraits,
    HasStrictTraits,
    Instance,
    List,
    Property,
    Str,
    cached_property,
    observe,
)
from traits.trait_base import SequenceTypes

from .editors.key_binding_editor import KeyBindingEditor
from .editors.list_editor import ListEditor
from .group import HGroup
from .handler import ModelView
from .item import Item
from .toolkit import toolkit
from .view import View


#: Trait definition for key bindings
Binding = Str(editor=KeyBindingEditor())


class KeyBinding(HasStrictTraits):
    """Binds one or two keystrokes to a method."""

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: First key binding
    binding1 = Binding

    #: Second key binding
    binding2 = Binding

    #: Description of what application function the method performs
    description = Str()

    #: Name of controller method the key is bound to
    method_name = Str()

    #: KeyBindings object that "owns" the KeyBinding
    owner = Instance("KeyBindings")

    def match(self, binding):
        return any(binding == x for x in {self.binding1, self.binding2})

    def clear_binding(self, binding):
        if binding == self.binding1:
            self.binding1 = self.binding2
            self.binding2 = ""
        if binding == self.binding2:
            self.binding2 = ""

    # -------------------------------------------------------------------------
    #  Traits view definitions:
    # -------------------------------------------------------------------------

    traits_view = View(
        HGroup(
            Item("binding1"),
            Item("binding2"),
            Item("description", style="readonly"),
            show_labels=False,
        )
    )


class KeyBindings(HasPrivateTraits):
    """A set of key bindings."""

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Set of defined key bindings (redefined dynamically)
    bindings = List(Instance(KeyBinding))

    #: Optional prefix to add to each method name
    prefix = Str()

    #: Optional suffix to add to each method name
    suffix = Str()

    # -- Private Traits -------------------------------------------------------

    #: The (optional) list of controllers associated with this KeyBindings
    #: object. The controllers may also be provided with the 'do' method:
    controllers = List(transient=True)

    #: The 'parent' KeyBindings object of this one (if any):
    parent = Instance("KeyBindings", transient=True)

    #: The root of the KeyBindings tree this object is part of:
    root = Property(observe="parent")

    #: The child KeyBindings of this object (if any):
    children = List(transient=True)

    # -------------------------------------------------------------------------
    #  Traits view definitions:
    # -------------------------------------------------------------------------

    def __init__(self, *bindings, **traits):
        # initialize bindings
        if len(bindings) == 1 and isinstance(bindings[0], SequenceTypes):
            bindings = bindings[0]
        traits.setdefault("bindings", list(bindings))
        for binding in traits["bindings"]:
            binding.owner = self

        super().__init__(**traits)

    def do(self, event, controllers=[], *args, **kw):
        """Processes a keyboard event."""
        if isinstance(controllers, dict):
            controllers = list(controllers.values())
        elif not isinstance(controllers, SequenceTypes):
            controllers = [controllers]
        else:
            controllers = list(controllers)

        return self._do(
            toolkit().key_event_to_name(event),
            controllers,
            args,
            kw.get("recursive", False),
        )

    def merge(self, key_bindings):
        """Merges another set of key bindings into this set."""
        binding_dic = {}
        for binding in self.bindings:
            binding_dic[binding.method_name] = binding

        for binding in key_bindings.bindings:
            binding2 = binding_dic.get(binding.method_name)
            if binding2 is not None:
                binding2.binding1 = binding.binding1
                binding2.binding2 = binding.binding2

    def clone(self, **traits):
        """Returns a clone of the KeyBindings object."""
        return self.__class__(*self.bindings, **traits).trait_set(
            **self.trait_get("prefix", "suffix")
        )

    def dispose(self):
        """Dispose of the object."""
        if self.parent is not None:
            self.parent.children.remove(self)

        del self.controllers
        del self.children
        del self.bindings

        self.parent = self._root = None

    def edit(self):
        """Edits a possibly hierarchical set of KeyBindings."""
        model_view = KeyBindingsHandler(model=self)
        ui = model_view.edit_traits()

    # -- Property Implementations ---------------------------------------------

    @cached_property
    def _get_root(self):
        root = self
        while root.parent is not None:
            root = root.parent

        return root

    # -- Event Handlers -------------------------------------------------------

    @observe('bindings:items:[binding1,binding2]')
    def _binding_updated(self, event):
        if event.new != "":
            for a_binding in self._match_binding(event.new, skip={event.object}):
                a_binding.clear_binding(event.new)

    @observe("children.items")
    def _children_modified(self, event):
        """Handles child KeyBindings being added to the object."""
        # the full children list is changed
        if isinstance(event.object, KeyBindings):
            for item in event.new:
                item.parent = self
        # the contents of the children list are changed
        else:
            for item in event.added:
                item.parent = self

    # -- Private Methods ------------------------------------------------------

    def _get_bindings(self, bindings):
        """Returns all of the bindings of this object and all of its children."""
        bindings.extend(self.bindings)
        for child in self.children:
            child._get_bindings(bindings)

        return bindings

    def _do(self, key_name, controllers, args, recursive):
        """Process the specified key for the specified set of controllers for
        this KeyBindings object and all of its children.
        """
        # Search through our own bindings for a match:
        for binding in self._match_binding(key_name):
            method_name = "%s%s%s" % (
                self.prefix,
                binding.method_name,
                self.suffix,
            )
            for controller in controllers + self.controllers:
                method = getattr(controller, method_name, None)
                if method is not None:
                    result = method(*args)
                    if result is not False:
                        return True

            if binding.method_name == "edit_bindings":
                self.edit()
                return True

        # If recursive, continue searching through a children's bindings:
        if recursive:
            for child in self.children:
                if child._do(key_name, controllers, args, recursive):
                    return True

        # Indicate no one processed the key:
        return False

    def _match_binding(self, binding, skip=frozenset()):
        """Return all KeyBinding instances that match the given binding.
        """
        return (
            a_binding
            for a_binding in self.bindings
            if a_binding not in skip and a_binding.match(binding)
        )


class KeyBindingsHandler(ModelView):

    bindings = List(Instance(KeyBinding))

    def key_binding_for(self, binding, key_name):
        """Returns the current binding for a specified key (if any)."""
        if key_name != "":
            for a_binding in self._match_binding(key_name, skip={binding}):
                return a_binding

        return None

    def _match_binding(self, binding, skip=frozenset()):
        """Return all KeyBinding instances that match the given binding.
        """
        return (
            a_binding
            for a_binding in self.bindings
            if a_binding not in skip and a_binding.match(binding)
        )

    def _bindings_default(self):
        bindings = list(set(self.model.root._get_bindings([])))
        bindings.sort(key=lambda x: (x.binding1[-1:], x.binding1))
        return bindings

    traits_view = View(
        [
            Item(
                "bindings",
                style="readonly",
                show_label=False,
                editor=ListEditor(style="custom"),
            ),
            "|{Click on an entry field, then press the key to "
            "assign. Double-click a field to clear it.}<>",
        ],
        title="Update Key Bindings",
        kind="livemodal",
        resizable=True,
        width=0.4,
        height=0.4,
    )



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from traits.api import Any, Dict, HasTraits, Str
from traitsui.api import Item, ShellEditor, View
from traitsui.testing.api import UITester
from traitsui.tests._tools import (
    BaseTestMixin,
    requires_toolkit,
    ToolkitName,
)


class ShellTest(HasTraits):
    locals_str = Str()
    locals_dict = Dict(Str, Any)


def get_str_view(share=False):
    return View(
        Item(
            "locals_str",
            editor=ShellEditor(share=share),
        )
    )


def get_dict_view(share=False):
    return View(
        Item(
            "locals_dict",
            editor=ShellEditor(share=share),
        )
    )


@requires_toolkit([ToolkitName.qt, ToolkitName.wx])
class TestShellEditor(BaseTestMixin, unittest.TestCase):
    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def smoke_test(self, locals_type, share):
        shell_test = ShellTest()
        tester = UITester()
        if locals_type == "str":
            with tester.create_ui(shell_test, dict(view=get_str_view(share))):
                pass
        else:
            with tester.create_ui(shell_test, dict(view=get_dict_view(share))):
                pass

    def test_no_share_dict(self):
        self.smoke_test("dict", False)

    def test_share_dict(self):
        self.smoke_test("dict", True)

    def test_no_share_str(self):
        self.smoke_test("str", False)

    def test_share_str(self):
        self.smoke_test("str", True)



# (C) Copyright 2008-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

import wx

from pyface.font import Font as PyfaceFont
from pyface.util.font_parser import simple_parser
from traits.api import HasTraits, TraitError

from ..font_trait import (
    WxFont, TraitsFont, create_traitsfont, font_families, font_styles,
    font_to_str, font_weights
)


class FontExample(HasTraits):

    font = WxFont()


class TestPyQtFont(unittest.TestCase):

    def test_create_traitsfont(self):
        expected_outcomes = {}
        expected_outcomes[""] = TraitsFont(
            10, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL
        )

        for weight, wx_weight in font_weights.items():
            expected_outcomes[weight] = TraitsFont(
                10, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL, wx_weight
            )

        for style, wx_style in font_styles.items():
            expected_outcomes[style] = TraitsFont(
                10, wx.FONTFAMILY_DEFAULT, wx_style, wx.FONTWEIGHT_NORMAL
            )

        expected_outcomes["underline"] = TraitsFont(
            10, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL, True
        )

        for size in ["18", "18 pt", "18 point"]:
            expected_outcomes[size] = TraitsFont(
                18, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL
            )

        for family, wx_family in font_families.items():
            expected_outcomes[family] = TraitsFont(
                10, wx_family, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL
            )

        expected_outcomes["Courier"] = TraitsFont(
            10, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL, False, "Courier",
        )
        expected_outcomes["Comic Sans"] = TraitsFont(
            10, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_NORMAL, wx.FONTWEIGHT_NORMAL, False, "Comic Sans",
        )
        expected_outcomes["18 pt Bold Oblique Underline Comic Sans script"] = TraitsFont(
            18, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_SLANT, wx.FONTWEIGHT_BOLD, True, "Comic Sans",
        )

        for name, expected in expected_outcomes.items():
            with self.subTest(name=name):
                result = create_traitsfont(name)

                # test we get expected font
                self.assertIsInstance(result, TraitsFont)
                self.assert_qfont_equal(result, expected)

                # round-trip trhough font_to_str
                result_2 = create_traitsfont(font_to_str(result))
                self.assert_qfont_equal(result, result_2)

    def test_create_traitsfont_wx_font(self):
        font = wx.Font(
            18, wx.FONTFAMILY_SCRIPT, wx.FONTSTYLE_SLANT, wx.FONTWEIGHT_BOLD, True, "Comic Sans",
        )
        traits_font = create_traitsfont(font)

        self.assertIsInstance(traits_font, TraitsFont)
        self.assert_qfont_equal(traits_font, font)

    def test_create_traitsfont_system_default(self):
        font = wx.SystemSettings.GetFont(wx.SYS_DEFAULT_GUI_FONT)
        traits_font = create_traitsfont(font)

        self.assertIsInstance(traits_font, TraitsFont)
        self.assert_qfont_equal(traits_font, font)

    def test_create_traitsfont_pyface_font(self):
        args = simple_parser("18 pt Bold Oblique Underline Courier")
        font = PyfaceFont(**args)
        traits_font = create_traitsfont(font)

        self.assertIsInstance(traits_font, TraitsFont)
        self.assert_qfont_equal(traits_font, font.to_toolkit())

    def test_font_trait_default(self):
        obj = FontExample()

        self.assertIsInstance(obj.font, TraitsFont)
        self.assert_qfont_equal(obj.font, wx.SystemSettings.GetFont(wx.SYS_DEFAULT_GUI_FONT))

    def test_font_trait_str(self):
        obj = FontExample(font="18 pt Bold Oblique Underline Comic Sans script")
        wx_font = TraitsFont(
            18, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_SLANT, wx.FONTWEIGHT_BOLD, True, "Comic Sans",
        )

        self.assertIsInstance(obj.font, TraitsFont)
        self.assert_qfont_equal(obj.font, wx_font)

    def test_font_trait_wx_font(self):
        wx_font = TraitsFont(
            18, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_SLANT, wx.FONTWEIGHT_BOLD, True, "Comic Sans",
        )
        obj = FontExample(font=wx_font)

        self.assertIsInstance(obj.font, TraitsFont)
        self.assert_qfont_equal(obj.font, wx_font)

    def test_font_trait_pyface_font(self):
        args = simple_parser("18 pt Bold Oblique Underline Courier typewriter")
        font = PyfaceFont(**args)
        obj = FontExample(font=font)

        self.assertIsInstance(obj.font, TraitsFont)
        self.assert_qfont_equal(obj.font, font.to_toolkit())

    def test_font_trait_none(self):
        obj = FontExample(font=None)

        self.assertIsNone(obj.font)

    def test_font_trait_bad(self):
        with self.assertRaises(TraitError):
            obj = FontExample(font=1)

    def test_traits_font_reduce(self):
        traits_font = TraitsFont(
            18, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_SLANT, wx.FONTWEIGHT_BOLD, True, "Comic Sans",
        )

        result = traits_font.__reduce_ex__(None)

        self.assertEqual(
            result,
            (
                create_traitsfont,
                ("18 point Comic Sans Oblique Bold underline",),
            ),
        )

    def test_traits_font_str(self):
        traits_font = TraitsFont(
            18, wx.FONTFAMILY_DEFAULT, wx.FONTSTYLE_SLANT, wx.FONTWEIGHT_BOLD, True, "Comic Sans",
        )

        result = str(traits_font)

        self.assertEqual(
            result,
            "18 point Comic Sans Oblique Bold underline",
        )

    def assert_wxfont_equal(self, font, other):
        self.assertIsInstance(font, wx.Font)
        self.assertIsInstance(other, wx.Font)

        self.assertEqual(font.FaceName, font.FaceName)
        self.assertEqual(font.Family, other.Family)
        self.assertEqual(font.PointSize, other.PointSize)
        self.assertEqual(font.Style, other.Style)
        self.assertEqual(font.Weight, other.Weight)
        self.assertEqual(font.GetUnderlined(), other.GetUnderlined())



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" General regression tests for various fixed bugs.
"""

from subprocess import check_output, STDOUT
import sys
import unittest

from traits.api import DelegatesTo, Event, HasTraits, Instance, Undefined
from traitsui.api import Editor, TextEditor

from traitsui.tests._tools import (
    BaseTestMixin,
)


class Parent(HasTraits):
    button = Event()


class Child(HasTraits):
    parent = Instance(Parent)
    button = DelegatesTo("parent")


class TestRegression(BaseTestMixin, unittest.TestCase):
    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def test_editor_on_delegates_to_event(self):
        """Make sure that DelegatesTo on Events passes Editor creation."""
        child = Child(parent=Parent())
        editor = Editor(
            None, factory=TextEditor(), object=child, name="button"
        )
        self.assertIs(editor.old_value, Undefined)

    def test_attribute_error(self):
        """Make sure genuine AttributeErrors raise on Editor creation."""
        self.assertRaises(
            AttributeError,
            Editor,
            None,
            factory=TextEditor(),
            object=Parent(),
            name="not_a_trait",
        )

    def test_importing_view_does_not_import_toolkit(self):
        output = check_output(
            [sys.executable, "-c",
             "import sys; from traitsui.view import View;"
             "print(list(sys.modules.keys()))"],
            stderr=STDOUT
        )
        output = output.decode('ascii')
        self.assertFalse('QtCore' in output)



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Test the initialization of a Datetime when given minimum and maximum
dates in the DatetimeEditor
"""
from datetime import datetime, timedelta
import unittest

from traits.has_traits import HasTraits
from traits.trait_types import Datetime
from traitsui.editors.datetime_editor import DatetimeEditor
from traitsui.item import Item
from traitsui.view import View


from traitsui.tests._tools import (
    BaseTestMixin,
    create_ui,
    requires_toolkit,
    reraise_exceptions,
    ToolkitName,
)

MAX_DATE = datetime.now()
MIN_DATE = MAX_DATE - timedelta(days=10)
DATE = MAX_DATE - timedelta(days=5)


class DatetimeInitDailog(HasTraits):

    date = Datetime(allow_none=True)

    def _date_default(self):
        return DATE

    traits_view = View(Item('date',
                            editor=DatetimeEditor(
                                maximum_datetime=MAX_DATE,
                                minimum_datetime=MIN_DATE)))


class TestDatetimeEditor(BaseTestMixin, unittest.TestCase):

    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    @requires_toolkit([ToolkitName.qt])
    def test_date_initialization(self):
        # Behavior: for a date initialized between maximum and minimum dates in
        # DatetimeEditor, that date should be unmodified

        dialog = DatetimeInitDailog()

        # verify that the date is correctly initialized
        self.assertEqual(dialog.date, DATE)

        with reraise_exceptions(), create_ui(dialog) as ui:
            # verify that the date hasn't changed after creating the ui
            self.assertEqual(dialog.date, DATE)



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""Demo of the ArrayEditor"""

import numpy as np

from traits.api import Array, HasPrivateTraits
from traitsui.api import ArrayEditor, Item, View
from traitsui.menu import NoButtons


class ArrayEditorTest(HasPrivateTraits):

    three = Array(np.int64, (3, 3))

    four = Array(
        np.float64,
        (4, 4),
        editor=ArrayEditor(width=-50),
    )

    view = View(
        Item('three', label='3x3 Integer'),
        '_',
        Item('three', label='Integer Read-only', style='readonly'),
        '_',
        Item('four', label='4x4 Float'),
        '_',
        Item('four', label='Float Read-only', style='readonly'),
        buttons=NoButtons,
        resizable=True,
    )


demo = ArrayEditorTest()

if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest
from unittest.mock import Mock

from traits.testing.api import UnittestTools


from ..key_bindings import KeyBinding, KeyBindings, KeyBindingsHandler


class Controller1:

    def __init__(self):
        self.called = None

    def m1(self, *args):
        self.called = ("m1", args)


class Controller2(Controller1):

    def m2(self, *args):
        self.called = ("m2", args)


class TestKeyBinding(UnittestTools, unittest.TestCase):

    def test_clear_binding_no_match(self):
        key_binding = KeyBinding(binding1="Ctrl-S", binding2="Ctrl-T")

        with self.assertTraitDoesNotChange(key_binding, "binding1"):
            with self.assertTraitDoesNotChange(key_binding, "binding2"):
                key_binding.clear_binding("Ctrl-U")

    def test_clear_binding_match_binding2(self):
        key_binding = KeyBinding(binding1="Ctrl-S", binding2="Ctrl-T")

        with self.assertTraitDoesNotChange(key_binding, "binding1"):
            with self.assertTraitChanges(key_binding, "binding2"):
                key_binding.clear_binding("Ctrl-T")

        self.assertEqual(key_binding.binding2, "")

    def test_clear_binding_match_binding1(self):
        key_binding = KeyBinding(binding1="Ctrl-S", binding2="Ctrl-T")

        with self.assertTraitChanges(key_binding, "binding1"):
            with self.assertTraitChanges(key_binding, "binding2"):
                key_binding.clear_binding("Ctrl-S")

        self.assertEqual(key_binding.binding1, "Ctrl-T")
        self.assertEqual(key_binding.binding2, "")

    def test_match(self):
        key_binding = KeyBinding(binding1="Ctrl-S", binding2="Ctrl-T")

        self.assertTrue(key_binding.match("Ctrl-S"))
        self.assertTrue(key_binding.match("Ctrl-T"))
        self.assertFalse(key_binding.match("Ctrl-U"))


class TestKeyBindings(UnittestTools, unittest.TestCase):

    def test_init(self):
        bindings = [
            KeyBinding(binding1="Ctrl-S", binding2="Ctrl-T"),
            KeyBinding(binding1="Ctrl-U", binding2=""),
        ]
        key_bindings = KeyBindings(bindings)

        self.assertEqual(key_bindings.bindings, bindings)

    def test_init_args(self):
        bindings = [
            KeyBinding(binding1="Ctrl-S", binding2="Ctrl-T"),
            KeyBinding(binding1="Ctrl-U", binding2=""),
        ]
        key_bindings = KeyBindings(*bindings)

        self.assertEqual(key_bindings.bindings, bindings)

    def test_init_kwargs(self):
        bindings = [
            KeyBinding(binding1="Ctrl-S", binding2="Ctrl-T"),
            KeyBinding(binding1="Ctrl-U", binding2=""),
        ]
        key_bindings = KeyBindings(bindings=bindings)

        self.assertEqual(key_bindings.bindings, bindings)

    def test_merge(self):
        bindings = [
            KeyBinding(binding1="Ctrl-S", binding2="Ctrl-T", method_name="m1"),
            KeyBinding(binding1="Ctrl-U", binding2="", method_name="m2"),
        ]
        key_bindings_1 = KeyBindings(*bindings)

        key_bindings_2 = KeyBindings(
            KeyBinding(binding1="Ctrl-V", binding2="Ctrl-W", method_name="m2"),
            KeyBinding(binding1="Ctrl-S", binding2="", method_name="m3"),
        )

        key_bindings_1.merge(key_bindings_2)

        self.assertEqual(key_bindings_1.bindings, bindings)
        key_binding = bindings[1]
        self.assertEqual(key_binding.binding1, "Ctrl-V")
        self.assertEqual(key_binding.binding2, "Ctrl-W")

    def test_clear_bindings_match(self):
        bindings = [
            KeyBinding(binding1="Ctrl-S", binding2="Ctrl-T", method_name="m1"),
            KeyBinding(binding1="Ctrl-U", binding2="", method_name="m2"),
        ]
        key_bindings = KeyBindings(bindings=bindings)

        with self.assertTraitChanges(bindings[0], "binding2"):
            bindings[1].binding2 = "Ctrl-T"

        self.assertIs(bindings[0].binding2, "")

    def test_clear_bindings_no_match(self):
        bindings = [
            KeyBinding(binding1="Ctrl-S", binding2="Ctrl-T", method_name="m1"),
            KeyBinding(binding1="Ctrl-U", binding2="", method_name="m2"),
        ]
        key_bindings = KeyBindings(bindings=bindings)

        with self.assertTraitDoesNotChange(bindings[0], "binding1"):
            with self.assertTraitDoesNotChange(bindings[0], "binding2"):
                bindings[1].binding2 = "Ctrl-V"

    def test_do_match(self):
        bindings = [
            KeyBinding(binding1="Ctrl-S", binding2="Ctrl-T", method_name="m1"),
            KeyBinding(binding1="Ctrl-U", binding2="", method_name="m2"),
        ]
        key_bindings = KeyBindings(bindings=bindings)

        controller = Controller1()

        # test _do rather than do since it doesn't need a toolkit event
        result = key_bindings._do("Ctrl-T", [controller], ("args",), False)

        self.assertTrue(result)
        self.assertEqual(controller.called, ("m1", ("args",)))

    def test_do_match_first(self):
        bindings = [
            KeyBinding(binding1="Ctrl-S", binding2="Ctrl-T", method_name="m1"),
            KeyBinding(binding1="Ctrl-U", binding2="", method_name="m2"),
        ]
        key_bindings = KeyBindings(bindings=bindings)

        controller_1 = Controller1()
        controller_2 = Controller2()

        # test _do rather than do since it doesn't need a toolkit event
        result = key_bindings._do(
            "Ctrl-T",
            [controller_1, controller_2],
            ("args",),
            False,
        )

        self.assertTrue(result)
        self.assertEqual(controller_1.called, ("m1", ("args",)))

    def test_do_match_second(self):
        bindings = [
            KeyBinding(binding1="Ctrl-S", binding2="Ctrl-T", method_name="m1"),
            KeyBinding(binding1="Ctrl-U", binding2="", method_name="m2"),
        ]
        key_bindings = KeyBindings(bindings=bindings)

        controller_1 = Controller1()
        controller_2 = Controller2()

        # test _do rather than do since it doesn't need a toolkit event
        result = key_bindings._do(
            "Ctrl-U",
            [controller_1, controller_2],
            ("args",),
            False,
        )

        self.assertTrue(result)
        self.assertEqual(controller_2.called, ("m2", ("args",)))

    def test_do_no_match(self):
        bindings = [
            KeyBinding(binding1="Ctrl-S", binding2="Ctrl-T", method_name="m1"),
            KeyBinding(binding1="Ctrl-U", binding2="", method_name="m2"),
        ]
        key_bindings = KeyBindings(bindings=bindings)

        controller = Controller1()

        # test _do rather than do since it doesn't need a toolkit event
        result = key_bindings._do("Ctrl-U", [controller], ("args",), False)

        self.assertFalse(result)
        self.assertIsNone(controller.called)

    def test_do_no_match_complete(self):
        bindings = [
            KeyBinding(binding1="Ctrl-S", binding2="Ctrl-T", method_name="m1"),
            KeyBinding(binding1="Ctrl-U", binding2="", method_name="m2"),
        ]
        key_bindings = KeyBindings(bindings=bindings)

        controller = Controller1()

        # test _do rather than do since it doesn't need a toolkit event
        result = key_bindings._do("Ctrl-V", [controller], ("args",), False)

        self.assertFalse(result)
        self.assertIsNone(controller.called)


class TestKeyBindingsHandler(unittest.TestCase):

    def test_key_binding_for_match(self):
        bindings = [
            KeyBinding(binding1="Ctrl-S", binding2="Ctrl-T", method_name="m1"),
            KeyBinding(binding1="Ctrl-U", binding2="", method_name="m2"),
        ]
        key_bindings = KeyBindings(bindings=bindings)
        handler = KeyBindingsHandler(model=key_bindings)

        binding = handler.key_binding_for(bindings[0], "Ctrl-U")

        self.assertIs(binding, bindings[1])

    def test_key_binding_for_no_match(self):
        bindings = [
            KeyBinding(binding1="Ctrl-S", binding2="Ctrl-T", method_name="m1"),
            KeyBinding(binding1="Ctrl-U", binding2="", method_name="m2"),
        ]
        key_bindings = KeyBindings(bindings=bindings)
        handler = KeyBindingsHandler(model=key_bindings)

        binding = handler.key_binding_for(bindings[0], "Ctrl-V")

        self.assertIsNone(binding)

    def test_key_binding_for_match_self(self):
        bindings = [
            KeyBinding(binding1="Ctrl-S", binding2="Ctrl-T", method_name="m1"),
            KeyBinding(binding1="Ctrl-U", binding2="", method_name="m2"),
        ]
        key_bindings = KeyBindings(bindings=bindings)
        handler = KeyBindingsHandler(model=key_bindings)

        binding = handler.key_binding_for(bindings[0], "Ctrl-S")

        self.assertIsNone(binding)

    def test_key_binding_for_match_empty(self):
        bindings = [
            KeyBinding(binding1="Ctrl-S", binding2="Ctrl-T", method_name="m1"),
            KeyBinding(binding1="Ctrl-U", binding2="", method_name="m2"),
        ]
        key_bindings = KeyBindings(bindings=bindings)
        handler = KeyBindingsHandler(model=key_bindings)

        binding = handler.key_binding_for(bindings[0], "")

        self.assertIsNone(binding)



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
This example demonstrates how to test the use of visible_when to make UI
components visible or not

The GUI being tested is written in the demo under the same name (minus the
preceding 'test') in the outer directory.
"""

import os
import runpy
import unittest

from traitsui.testing.api import IsVisible, UITester

#: Filename of the demo script
FILENAME = "visible_when.py"

#: Path of the demo script
DEMO_PATH = os.path.join(os.path.dirname(__file__), "..", FILENAME)


class TestVisibleWhenDemo(unittest.TestCase):
    def test_visible_when_demo(self):
        demo = runpy.run_path(DEMO_PATH)["demo"]

        tester = UITester()
        with tester.create_ui(demo) as ui:
            # person should be 16 by default
            self.assertLess(demo.age, 18)

            marital_status_field = tester.find_by_name(ui, 'marital_status')
            legal_guardian_field = tester.find_by_name(ui, 'legal_guardian')
            self.assertFalse(marital_status_field.inspect(IsVisible()))
            self.assertTrue(legal_guardian_field.inspect(IsVisible()))

            # UITester is yet to support SimpleSpinEditor so we change the
            # value of the trait directly
            demo.age = 20

            registered_voter_field = tester.find_by_name(
                ui, 'registered_voter'
            )
            school_field = tester.find_by_name(ui, 'school')
            self.assertTrue(registered_voter_field.inspect(IsVisible()))
            self.assertFalse(school_field.inspect(IsVisible()))


# Run the test(s)
unittest.TextTestRunner().run(
    unittest.TestLoader().loadTestsFromTestCase(TestVisibleWhenDemo)
)



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Editor that displays an interactive Python shell.
"""

from traits.api import Bool, Str, Event, Property, observe
from traits.observation.api import match

from traitsui.basic_editor_factory import BasicEditorFactory
from traitsui.editor import Editor
from traitsui.toolkit import toolkit_object


class _ShellEditor(Editor):
    """Base class for an editor that displays an interactive Python shell."""

    #: An event fired to execute a command in the shell.
    command_to_execute = Event()

    #: An event fired whenver the user executes a command in the shell:
    command_executed = Event(Bool)

    #: Is the shell editor is scrollable? This value overrides the default.
    scrollable = True

    # -------------------------------------------------------------------------
    # 'Editor' Interface
    # -------------------------------------------------------------------------

    def init(self, parent):
        """Finishes initializing the editor by creating the underlying toolkit
        widget.
        """
        # Moving the import here, since PythonShell is implemented in the
        # Pyface backend packages, and we want to delay loading this toolkit
        # specific class until this editor is actually used.
        from pyface.python_shell import PythonShell

        locals = None
        self._base_locals = None
        value = self.value
        if self.factory.share and isinstance(value, dict):
            locals = value
        self._shell = shell = PythonShell(parent)
        self.control = shell.control
        if locals:
            for item in locals.items():
                shell.bind(*item)
        if locals is None:
            object = self.object
            shell.bind("self", object)
            shell.observe(
                self.update_object, "command_executed", dispatch="ui"
            )
            if not isinstance(value, dict):
                self._any_trait_observer = lambda name, ctrait: True
                object.observe(
                    self.update_any,
                    match(self._any_trait_observer),
                    dispatch="ui",
                )
            else:
                self._base_locals = locals = {}
                for name in self._shell.interpreter().locals.keys():
                    locals[name] = None

        # Synchronize any editor events:
        self.sync_value(
            self.factory.command_to_execute, "command_to_execute", "from"
        )
        self.sync_value(
            self.factory.command_executed, "command_executed", "to"
        )

        self.set_tooltip()

    def update_object(self, event):
        """Handles the user entering input data in the edit control."""
        locals = self._shell.interpreter().locals
        base_locals = self._base_locals
        if base_locals is None:
            object = self.object
            for name in object.trait_names():
                if name in locals:
                    try:
                        setattr(object, name, locals[name])
                    except:
                        pass
        else:
            dic = self.value
            for name in locals.keys():
                if name not in base_locals:
                    try:
                        dic[name] = locals[name]
                    except:
                        pass

        self.command_executed = True

    def update_editor(self):
        """Updates the editor when the object trait changes externally to the
        editor.
        """
        if self.factory.share:
            value = self.value
            if isinstance(value, dict):
                self._shell.interpreter().locals = value
        else:
            locals = self._shell.interpreter().locals
            base_locals = self._base_locals
            if base_locals is None:
                object = self.object
                for name in object.trait_names():
                    locals[name] = getattr(object, name, None)
            else:
                dic = self.value
                for name, value in dic.items():
                    locals[name] = value

    def update_any(self, event):
        """Updates the editor when the object trait changes externally to the
        editor.
        """
        name, new = event.name, event.new
        locals = self._shell.interpreter().locals
        if self._base_locals is None:
            locals[name] = new
        else:
            self.value[name] = new

    def dispose(self):
        """Disposes of the contents of an editor."""
        if not (self.factory.share and isinstance(self.value, dict)):
            self._shell.observe(
                self.update_object,
                "command_executed",
                remove=True,
                dispatch="ui",
            )
            if self._base_locals is None:
                self.object.observe(
                    self.update_any,
                    match(self._any_trait_observer),
                    remove=True,
                    dispatch="ui",
                )

        super().dispose()

    def restore_prefs(self, prefs):
        """Restores any saved user preference information associated with the
        editor.
        """
        shell = self._shell
        try:
            history = prefs.get("history", [])
            history_index = prefs.get("history_index", -1)
            shell.set_history(history, history_index)
        except:
            pass

    def save_prefs(self):
        """Returns any user preference information associated with the editor."""
        history, history_index = self._shell.get_history()
        return {"history": history, "history_index": history_index}

    # -------------------------------------------------------------------------
    # Private Interface
    # -------------------------------------------------------------------------

    # Trait change handlers --------------------------------------------------

    @observe("command_to_execute")
    def _execute_command(self, event):
        """Handles the 'command_to_execute' trait being fired."""
        # Show the command. A 'hidden' command should be executed directly on
        # the namespace trait!
        command = event.new
        self._shell.execute_command(command, hidden=False)


class ShellEditor(BasicEditorFactory):
    """Editor factory for shell editors."""

    #: The editor class to be instantiated.
    klass = Property()

    #: Should the shell interpreter use the object value's dictionary?
    share = Bool(False)

    #: Extended trait name of the object event trait which triggers a command
    #: execution in the shell when fired.
    command_to_execute = Str()

    #: Extended trait name of the object event trait which is fired when a
    #: command is executed.
    command_executed = Str()

    def _get_klass(self):
        """Returns the toolkit-specific editor class to be used in the UI."""
        return toolkit_object("shell_editor:_ShellEditor")


# This alias is deprecated and will be removed in TraitsUI 8.
ToolkitEditorFactory = ShellEditor



# (C) Copyright 2008-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# ------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD license.
# However, when used with the GPL version of PyQt the additional terms
# described in the PyQt GPL exception also apply

#
# Author: Riverbank Computing Limited
# ------------------------------------------------------------------------------

""" Defines constants used by the PyQt implementation of the various text
editors and text editor factories.
"""


from pyface.qt import QtGui

from pyface.api import SystemMetrics

_palette = QtGui.QApplication.palette()

_base = _palette.color(QtGui.QPalette.ColorRole.Base)

#: We are in dark mode if field background colour HSV value is dark.
is_dark = (_base.value() < 0x80)

#: Default dialog title
DefaultTitle = "Edit properties"

#: Color of valid input
OKColor = _base

#: Color to highlight input errors
if is_dark:
    ErrorColor = QtGui.QColor(0xcf, 0x66, 0x79)
else:
    ErrorColor = QtGui.QColor(255, 192, 192)

#: Color for background of read-only fields
ReadonlyColor = _palette.color(QtGui.QPalette.ColorRole.Window)

#: Color for background of fields where objects can be dropped
DropColor = _palette.color(QtGui.QPalette.ColorRole.Base)

#: Color for an editable field
EditableColor = _base

#: Color for background of windows (like dialog background color)
WindowColor = _palette.color(QtGui.QPalette.ColorRole.Window)

del _palette

#: Screen width
screen_dx = SystemMetrics().screen_width

#: Screen height
screen_dy = SystemMetrics().screen_height



# (C) Copyright 2008-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD license.
# However, when used with the GPL version of PyQt the additional terms
# described in the PyQt GPL exception also apply
#
# Author: Riverbank Computing Limited

""" Defines various directory editor for the PyQt user interface toolkit.
"""

from pyface.api import DirectoryDialog

from .file_editor import (
    SimpleEditor as SimpleFileEditor,
    CustomEditor as CustomFileEditor,
)


class SimpleEditor(SimpleFileEditor):
    """Simple style of editor for directories, which displays a text field
    and a **Browse** button that opens a directory-selection dialog box.
    """

    def _create_file_dialog(self):
        """Creates the correct type of file dialog."""

        dlg = DirectoryDialog(
            parent=self.get_control_widget(),
            default_path=self._file_name.text(),
        )
        return dlg


class CustomEditor(CustomFileEditor):
    """Custom style of editor for directories, which displays a tree view of
    the file system.
    """

    def init(self, parent):
        super().init(parent)
        self._model.setNameFilterDisables(True)
        self._model.setNameFilters([""])

    def update_object(self, idx):
        """Handles the user changing the contents of the edit control."""
        if self.control is not None:
            if self._model.isDir(idx):
                self.value = str(self._model.filePath(idx))

    # Trait change handlers --------------------------------------------------

    def _filter_changed(self):
        """Handles the 'filter' trait being changed."""
        # name filters don't apply to directories
        pass



# (C) Copyright 2009-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# ------------------------------------------------------------------------------
# Copyright (c) 2008, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD license.
# However, when used with the GPL version of PyQt the additional terms
# described in the PyQt GPL exception also apply

#
# Author: Riverbank Computing Limited
# ------------------------------------------------------------------------------

""" Defines the various instance editors and the instance editor factory for
    the PyQt user interface toolkit..
"""


from pyface.qt import QtCore, QtGui

from traits.api import HasTraits, Instance, Property
from traits.observation.api import trait

from traitsui.ui_traits import AView
from traitsui.helper import user_name_for
from traitsui.handler import Handler
from traitsui.instance_choice import InstanceChoiceItem
from .editor import Editor
from .drop_editor import _DropEventFilter
from .constants import DropColor
from .helper import position_window


OrientationMap = {
    "default": None,
    "horizontal": QtGui.QBoxLayout.Direction.LeftToRight,
    "vertical": QtGui.QBoxLayout.Direction.TopToBottom,
}


class CustomEditor(Editor):
    """Custom style of editor for instances. If selection among instances is
    allowed, the editor displays a combo box listing instances that can be
    selected. If the current instance is editable, the editor displays a panel
    containing trait editors for all the instance's traits.
    """

    #: Background color when an item can be dropped on the editor:
    ok_color = DropColor

    #: The orientation of the instance editor relative to the instance selector:
    orientation = QtGui.QBoxLayout.Direction.TopToBottom

    #: Class constant:
    extra = 0

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: List of InstanceChoiceItem objects used by the editor
    items = Property()

    #: The view to use for displaying the instance
    view = AView

    def init(self, parent):
        """Finishes initializing the editor by creating the underlying toolkit
        widget.
        """
        factory = self.factory
        if factory.name != "":
            self._object, self._name, self._value = self.parse_extended_name(
                factory.name
            )

        # Create a panel to hold the object trait's view:
        if factory.editable:
            self.control = self._panel = parent = QtGui.QWidget()

        # Build the instance selector if needed:
        selectable = factory.selectable
        droppable = factory.droppable
        items = self.items
        for item in items:
            droppable |= item.is_droppable()
            selectable |= item.is_selectable()

        if selectable:
            self._object_cache = {}
            item = self.item_for(self.value)
            if item is not None:
                self._object_cache[id(item)] = self.value

            self._choice = QtGui.QComboBox()
            self._choice.activated.connect(self.update_object)

            self.set_tooltip(self._choice)

            if factory.name != "":
                self._object.observe(
                    self.rebuild_items, self._name + ".items", dispatch="ui"
                )

            factory.observe(self.rebuild_items, "values.items", dispatch="ui")

            self.rebuild_items()

        elif droppable:
            self._choice = QtGui.QLineEdit()
            self._choice.setReadOnly(True)
            self.set_tooltip(self._choice)

        orientation = OrientationMap[factory.orientation]
        if orientation is None:
            orientation = self.orientation

        if (selectable or droppable) and factory.editable:
            layout = QtGui.QBoxLayout(orientation, parent)
            layout.setContentsMargins(0, 0, 0, 0)
            layout.addWidget(self._choice)

            if orientation == QtGui.QBoxLayout.Direction.TopToBottom:
                hline = QtGui.QFrame()
                hline.setFrameShape(QtGui.QFrame.Shape.HLine)
                hline.setFrameShadow(QtGui.QFrame.Shadow.Sunken)

                layout.addWidget(hline)

            self.create_editor(parent, layout)
        elif self.control is None:
            if self._choice is None:
                self._choice = QtGui.QComboBox()
                self._choice.activated[int].connect(self.update_object)

            self.control = self._choice
        else:
            layout = QtGui.QBoxLayout(orientation, parent)
            layout.setContentsMargins(0, 0, 0, 0)
            self.create_editor(parent, layout)

        if droppable:
            # Install EventFilter on control to handle DND events.
            drop_event_filter = _DropEventFilter(self.control)
            self.control.installEventFilter(drop_event_filter)

        # Synchronize the 'view' to use:
        # fixme: A normal assignment can cause a crash (for unknown reasons) in
        # some cases, so we make sure that no notifications are generated:
        self.trait_setq(view=factory.view)
        self.sync_value(factory.view_name, "view", "from")

    def create_editor(self, parent, layout):
        """Creates the editor control."""
        self._panel = QtGui.QWidget()
        layout.addWidget(self._panel)

    def _get_items(self):
        """Gets the current list of InstanceChoiceItem items."""
        if self._items is not None:
            return self._items

        factory = self.factory
        if self._value is not None:
            values = self._value() + factory.values
        else:
            values = factory.values

        items = []
        adapter = factory.adapter
        for value in values:
            if not isinstance(value, InstanceChoiceItem):
                value = adapter(object=value)
            # rebuild_items when an item's name changes so it is reflected by
            # combobox. This change was added to fix enthought/traitsui#1641
            value.object.observe(
                self.rebuild_items,
                trait(value.name_trait, optional=True),
                dispatch="ui",
            )
            items.append(value)

        self._items = items

        return items

    def rebuild_items(self, event=None):
        """Rebuilds the object selector list."""
        # Clear the current cached values:
        self._items = None

        # Rebuild the contents of the selector list:
        name = -1
        value = self.value
        choice = self._choice
        choice.clear()
        for i, item in enumerate(self.items):
            if item.is_selectable():
                choice.addItem(item.get_name())
                if item.is_compatible(value):
                    name = i

        # Reselect the current item if possible:
        if name >= 0:
            choice.setCurrentIndex(name)
        else:
            # Otherwise, current value is no longer valid, set combobox empty
            try:
                self.value = None
                choice.setCurrentIndex(-1)
            except:
                pass

    def item_for(self, object):
        """Returns the InstanceChoiceItem for a specified object."""
        for item in self.items:
            if item.is_compatible(object):
                return item

        return None

    def view_for(self, object, item):
        """Returns the view to use for a specified object."""
        view = ""
        if item is not None:
            view = item.get_view()

        if view == "":
            view = self.view

        return self.ui.handler.trait_view_for(
            self.ui.info, view, object, self.object_name, self.name
        )

    def update_object(self, index):
        """Handles the user selecting a new value from the combo box."""
        item = self.items[index]
        id_item = id(item)
        object = self._object_cache.get(id_item)
        if object is None:
            object = item.get_object()
            if (not self.factory.editable) and item.is_factory:
                view = self.view_for(object, self.item_for(object))
                view.ui(object, self.control, "modal")

            if self.factory.cachable:
                self._object_cache[id_item] = object

        self.value = object
        self.resynch_editor()

    def update_editor(self):
        """Updates the editor when the object trait changes externally to the
        editor.
        """
        # Synchronize the editor contents:
        self.resynch_editor()

        # Update the selector (if any):
        choice = self._choice
        item = self.item_for(self.value)
        if choice is not None:
            if item is not None:
                name = item.get_name(self.value)
                if self._object_cache is not None:
                    idx = choice.findText(name)
                    if idx < 0:
                        idx = choice.count()
                        choice.addItem(name)

                    choice.setCurrentIndex(idx)
                else:
                    choice.setText(name)
            else:
                # choice can also be a QLineEdit in which case we just leave
                # text as empty
                if isinstance(choice, QtGui.QComboBox):
                    choice.setCurrentIndex(-1)

    def resynch_editor(self):
        """Resynchronizes the contents of the editor when the object trait
        changes externally to the editor.
        """
        panel = self._panel
        if panel is not None:
            # Dispose of the previous contents of the panel:
            layout = panel.layout()
            if layout is None:
                layout = QtGui.QVBoxLayout(panel)
                layout.setContentsMargins(0, 0, 0, 0)
            elif self._ui is not None:
                self._ui.dispose()
                self._ui = None
            else:
                child = layout.takeAt(0)
                while child is not None:
                    child = layout.takeAt(0)

                del child

            # Create the new content for the panel:
            stretch = 0
            value = self.value
            if not isinstance(value, HasTraits):
                str_value = ""
                if value is not None:
                    str_value = self.str_value
                control = QtGui.QLabel(str_value)
            else:
                view = self.view_for(value, self.item_for(value))
                context = value.trait_context()
                handler = None
                if isinstance(value, Handler):
                    handler = value
                context.setdefault("context", self.object)
                context.setdefault("context_handler", self.ui.handler)
                self._ui = ui = view.ui(
                    context,
                    panel,
                    "subpanel",
                    value.trait_view_elements(),
                    handler,
                    self.factory.id,
                )
                control = ui.control
                self.scrollable = ui._scrollable
                ui.parent = self.ui

                if view.resizable or view.scrollable or ui._scrollable:
                    stretch = 1

            # FIXME: Handle stretch.
            layout.addWidget(control)

    def dispose(self):
        """Disposes of the contents of an editor."""
        # Make sure we aren't hanging on to any object refs:
        self._object_cache = None

        if self._ui is not None:
            self._ui.dispose()

        if self._choice is not None:
            # _choice can also be a QLineEdit in which case we never set up
            # this observer.
            if isinstance(self._choice, QtGui.QComboBox):
                if self._object is not None:
                    self._object.observe(
                        self.rebuild_items,
                        self._name + ".items",
                        remove=True,
                        dispatch="ui",
                    )

                self.factory.observe(
                    self.rebuild_items,
                    "values.items",
                    remove=True,
                    dispatch="ui",
                )

        super().dispose()

    def error(self, excp):
        """Handles an error that occurs while setting the object's trait value."""
        pass

    def get_error_control(self):
        """Returns the editor's control for indicating error status."""
        return self._choice or self.control

    # -- UI preference save/restore interface ---------------------------------

    def restore_prefs(self, prefs):
        """Restores any saved user preference information associated with the
        editor.
        """
        ui = self._ui
        if (ui is not None) and (prefs.get("id") == ui.id):
            ui.set_prefs(prefs.get("prefs"))

    def save_prefs(self):
        """Returns any user preference information associated with the editor."""
        ui = self._ui
        if (ui is not None) and (ui.id != ""):
            return {"id": ui.id, "prefs": ui.get_prefs()}

        return None

    # -- Traits event handlers ------------------------------------------------

    def _view_changed(self, view):
        self.resynch_editor()


class SimpleEditor(CustomEditor):
    """Simple style of editor for instances, which displays a button. Clicking
    the button displays a dialog box in which the instance can be edited.
    """

    #: The ui instance for the currently open editor dialog
    _dialog_ui = Instance("traitsui.ui.UI")

    #: Class constants:
    orientation = QtGui.QBoxLayout.Direction.LeftToRight
    extra = 2

    def create_editor(self, parent, layout):
        """Creates the editor control (a button)."""
        self._button = QtGui.QPushButton()
        self._button.setAutoDefault(False)
        layout.addWidget(self._button)
        self._button.clicked.connect(self.edit_instance)
        # Make sure the editor is properly disposed if parent UI is closed
        self._button.destroyed.connect(self._parent_closed)

    def edit_instance(self):
        """Edit the contents of the object trait when the user clicks the
        button.
        """
        # Create the user interface:
        factory = self.factory
        view = self.ui.handler.trait_view_for(
            self.ui.info, factory.view, self.value, self.object_name, self.name
        )
        self._dialog_ui = self.value.edit_traits(
            view, kind=factory.kind, id=factory.id
        )

        # Check to see if the view was 'modal', in which case it will already
        # have been closed (i.e. is None) by the time we get control back:
        if self._dialog_ui.control is not None:
            # Position the window on the display:
            position_window(self._dialog_ui.control)

            # Chain our undo history to the new user interface if it does not
            # have its own:
            if self._dialog_ui.history is None:
                self._dialog_ui.history = self.ui.history

        else:
            self._dialog_ui = None

    def resynch_editor(self):
        """Resynchronizes the contents of the editor when the object trait
        changes externally to the editor.
        """
        button = self._button
        if button is not None:
            label = self.factory.label
            if label == "":
                label = user_name_for(self.name)

            button.setText(label)
            button.setEnabled(isinstance(self.value, HasTraits))

    def _parent_closed(self):
        if self._dialog_ui is not None:
            if self._dialog_ui.control is not None:
                self._dialog_ui.control.close()
            self._dialog_ui.dispose()
            self._dialog_ui = None



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Table column object for RGBColor traits.
"""

from traitsui.table_column import ObjectColumn


class ColorColumn(ObjectColumn):
    """Table column object for RGBColor traits."""

    #: For display by default.
    style = "readonly"

    # -- ObjectColumn Overrides -----------------------------------------------

    def get_cell_color(self, object):
        """Returns the cell background color for the column for a specified
        object.
        """
        color_values = getattr(object, self.name + "_", None)
        if color_values is None:
            tk_color = super().get_cell_color(object)
        elif isinstance(color_values, tuple):
            tk_color = self._as_int_rgb_tuple(color_values)
        else:
            tk_color = color_values
        return tk_color

    def get_value(self, object):
        """Gets the value of the column for a specified object."""
        value = getattr(self.get_object(object), self.name, "")
        if isinstance(value, tuple):
            value = self._float_rgb_tuple_to_str(value)
        elif not isinstance(value, str):
            value = ""
        return value

    # -- Private Methods ------------------------------------------------------

    def _as_int_rgb_tuple(self, color_values):
        """Returns object color as RGB integers."""
        return tuple(int(255 * v + 0.5) for v in color_values)

    def _float_rgb_tuple_to_str(self, color_values):
        """Returns object color as RGB floats."""
        csv = ", ".join("{:5.3f}".format(x) for x in color_values)
        return "({})".format(csv)



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Defining column-specific context menu in a Table

Shows a TableEditor which has column-specific context menus.

The demo is a simple baseball scoring system, which lists each player and
their current batting statistics. After a given player has an at bat, you
right-click on the table cell corresponding to the player and the result of
the at-bat (e.g. 'S' = single) and select the 'Add' menu option to register
that the player hit a single and update the player's overall statistics.

This demo also illustrates the use of Property traits, and how using 'event'
meta-data can simplify event handling by collapsing an event that can
occur on a number of traits into a category of event, which can be handled by
a single event handler defined for the category (in this case, the category
is 'affects_average').
"""

from random import randint
from traits.api import HasStrictTraits, Str, Int, Float, List, Property
from traitsui.api import Action, Item, Menu, ObjectColumn, TableEditor, View


# Define a custom table column for handling items which affect the player's
# batting average:
class AffectsAverageColumn(ObjectColumn):

    # Define the context menu for the column:
    menu = Menu(
        Action(name='Add', action='column.add(object)'),
        Action(name='Sub', action='column.sub(object)'),
    )

    # Right-align numeric values (override):
    horizontal_alignment = 'center'

    # Column width (override):
    width = 0.09

    # Don't allow the data to be edited directly:
    editable = False

    # Action methods for the context menu items:

    def add(self, object):
        """Increment the affected player statistic."""
        setattr(object, self.name, getattr(object, self.name) + 1)

    def sub(self, object):
        """Decrement the affected player statistic."""
        setattr(object, self.name, getattr(object, self.name) - 1)


# The 'players' trait table editor:
player_editor = TableEditor(
    editable=True,
    sortable=False,
    auto_size=False,
    columns=[
        ObjectColumn(name='name', editable=False, width=0.28),
        AffectsAverageColumn(name='at_bats', label='AB'),
        AffectsAverageColumn(name='strike_outs', label='SO'),
        AffectsAverageColumn(name='singles', label='S'),
        AffectsAverageColumn(name='doubles', label='D'),
        AffectsAverageColumn(name='triples', label='T'),
        AffectsAverageColumn(name='home_runs', label='HR'),
        AffectsAverageColumn(name='walks', label='W'),
        ObjectColumn(
            name='average',
            label='Ave',
            editable=False,
            width=0.09,
            horizontal_alignment='center',
            format='%0.3f',
        ),
    ],
)


# 'Player' class:
class Player(HasStrictTraits):

    # Trait definitions:
    name = Str()
    at_bats = Int()
    strike_outs = Int(event='affects_average')
    singles = Int(event='affects_average')
    doubles = Int(event='affects_average')
    triples = Int(event='affects_average')
    home_runs = Int(event='affects_average')
    walks = Int()
    average = Property(Float)

    def _get_average(self):
        """Computes the player's batting average from the current statistics."""
        if self.at_bats == 0:
            return 0.0

        return (
            float(self.singles + self.doubles + self.triples + self.home_runs)
            / self.at_bats
        )

    def _affects_average_changed(self):
        """Handles an event that affects the player's batting average."""
        self.at_bats += 1


class Team(HasStrictTraits):

    # Trait definitions:
    players = List(Player)

    # Trait view definitions:
    traits_view = View(
        Item('players', show_label=False, editor=player_editor),
        title='Baseball Scoring Demo',
        width=0.5,
        height=0.5,
        resizable=True,
    )


def random_player(name):
    """Generates and returns a random player."""
    p = Player(
        name=name,
        strike_outs=randint(0, 50),
        singles=randint(0, 50),
        doubles=randint(0, 20),
        triples=randint(0, 5),
        home_runs=randint(0, 30),
        walks=randint(0, 50),
    )
    return p.trait_set(
        at_bats=(
            p.strike_outs
            + p.singles
            + p.doubles
            + p.triples
            + p.home_runs
            + randint(100, 200)
        )
    )


# Create the demo:
demo = Team(
    players=[
        random_player(name)
        for name in [
            'Dave',
            'Mike',
            'Joe',
            'Tom',
            'Dick',
            'Harry',
            'Dirk',
            'Fields',
            'Stretch',
        ]
    ]
)

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest
from unittest.mock import Mock

from traits.api import HasTraits, Instance, Int, List, Str, Tuple

from traitsui.api import (
    Action,
    EvalTableFilter,
    Item,
    ObjectColumn,
    TableEditor,
    View,
)
from traitsui.tests._tools import (
    BaseTestMixin,
    requires_toolkit,
    ToolkitName,
)
from traitsui.testing.api import (
    Cell,
    Disabled,
    DisplayedText,
    KeySequence,
    KeyClick,
    MouseClick,
    MouseDClick,
    Selected,
    SelectedIndices,
    UITester,
)


class ListItem(HasTraits):
    """Items to visualize in a table editor"""

    value = Str()
    other_value = Int()


class ObjectListWithSelection(HasTraits):
    values = List(Instance(ListItem))
    selected = Instance(ListItem)
    selections = List(Instance(ListItem))
    selected_index = Int()
    selected_indices = List(Int)
    selected_column = Str()
    selected_columns = List(Str)
    selected_cell = Tuple(Instance(ListItem), Str)
    selected_cells = List(Tuple(Instance(ListItem), Str))
    selected_cell_index = Tuple(Int, Int)
    selected_cell_indices = List(Tuple(Int, Int))


class ObjectList(HasTraits):
    values = List(Instance(ListItem))


simple_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ]
        ),
    ),
    buttons=["OK"],
)

filtered_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            sortable=False,
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            filter=EvalTableFilter(expression="other_value > 4"),
        ),
    ),
    buttons=["OK"],
)

select_row_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            sortable=False,
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="row",
            selected="selected",
        ),
    ),
    buttons=["OK"],
)

select_rows_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            sortable=False,
            columns=[ObjectColumn(name="value")],
            selection_mode="rows",
            selected="selections",
        ),
    ),
    buttons=["OK"],
)

select_row_index_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            sortable=False,
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="row",
            selected_indices="selected_index",
        ),
    ),
    buttons=["OK"],
)

select_row_indices_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            sortable=False,
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="rows",
            selected_indices="selected_indices",
        ),
    ),
    buttons=["OK"],
)

select_column_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            sortable=False,
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="column",
            selected="selected_column",
        ),
    ),
    buttons=["OK"],
)

select_columns_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            sortable=False,
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="columns",
            selected="selected_columns",
        ),
    ),
    buttons=["OK"],
)

select_column_index_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            sortable=False,
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="column",
            selected_indices="selected_index",
        ),
    ),
    buttons=["OK"],
)

select_column_indices_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            sortable=False,
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="columns",
            selected_indices="selected_indices",
        ),
    ),
    buttons=["OK"],
)

select_cell_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            sortable=False,
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="cell",
            selected="selected_cell",
        ),
    ),
    buttons=["OK"],
)

select_cells_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            sortable=False,
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="cells",
            selected="selected_cells",
        ),
    ),
    buttons=["OK"],
)

select_cell_index_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            sortable=False,
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="cell",
            selected_indices="selected_cell_index",
        ),
    ),
    buttons=["OK"],
)

select_cell_indices_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            sortable=False,
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="cells",
            selected_indices="selected_cell_indices",
        ),
    ),
    buttons=["OK"],
)

edit_on_first_click_false_view = View(
    Item(
        "values",
        show_label=False,
        editor=TableEditor(
            sortable=False,
            columns=[
                ObjectColumn(name="value"),
                ObjectColumn(name="other_value"),
            ],
            selection_mode="row",
            selected="selected",
            edit_on_first_click=False
        ),
    ),
    buttons=["OK"],
)


@requires_toolkit([ToolkitName.qt])
class TestTableEditor(BaseTestMixin, unittest.TestCase):
    def setUp(self):
        BaseTestMixin.setUp(self)

    def tearDown(self):
        BaseTestMixin.tearDown(self)

    def test_table_editor(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        tester = UITester()
        with tester.create_ui(object_list, dict(view=simple_view)):
            pass

    def test_filtered_table_editor(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(other_value=i ** 2) for i in range(10)]
        )
        tester = UITester()
        with tester.create_ui(object_list, dict(view=filtered_view)) as ui:
            values = tester.find_by_name(ui, "values")
            filter = values._target.filter
            num_filtered_indices = len(values._target.filtered_indices)
            self.assertIsNotNone(filter)
            self.assertEqual(num_filtered_indices, 7)

    def test_table_editor_select_row(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )

        tester = UITester()
        with tester.create_ui(object_list, dict(view=select_row_view)) as ui:
            # click the first cell in the 6th row to select the row
            values_table = tester.find_by_name(ui, "values")
            row6_cell = values_table.locate(Cell(5, 0))
            row6_cell.perform(MouseClick())
            selected = values_table.inspect(Selected())

        self.assertEqual(selected, [object_list.values[5]])
        self.assertIs(selected[0], object_list.values[5])
        self.assertIs(object_list.selected, selected[0])

    def test_table_editor_select_rows(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        object_list.selections = object_list.values[5:7]

        tester = UITester()
        with tester.create_ui(object_list, dict(view=select_rows_view)) as ui:
            values_table = tester.find_by_name(ui, "values")
            selected = values_table.inspect(Selected())

        self.assertEqual(selected, object_list.values[5:7])

    def test_table_editor_select_row_index(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        object_list.selected_index = 5

        tester = UITester()
        with tester.create_ui(object_list, dict(view=select_row_index_view)) \
                as ui:
            values_table = tester.find_by_name(ui, "values")
            values_table.locate(Cell(5, 0)).perform(MouseClick())
            selected = values_table.inspect(SelectedIndices())

        self.assertEqual(selected, [5])

    def test_table_editor_select_row_indices(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        object_list.selected_indices = [5, 7, 8]

        view = select_row_indices_view
        tester = UITester()
        with tester.create_ui(object_list, dict(view=view)) as ui:
            values_table = tester.find_by_name(ui, "values")
            selected = values_table.inspect(SelectedIndices())

        self.assertEqual(selected, [5, 7, 8])

    def test_table_editor_select_column(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )

        tester = UITester()
        with tester.create_ui(object_list, dict(view=select_column_view)) \
                as ui:
            values_table = tester.find_by_name(ui, "values")
            # click a cell in the first column (the "value" column)
            first_cell = values_table.locate(Cell(0, 0))
            first_cell.perform(MouseClick())

            selected = values_table.inspect(Selected())

        self.assertEqual(selected, ["value"])
        self.assertIs(selected[0], "value")
        self.assertIs(selected[0], object_list.selected_column)

    def test_table_editor_select_columns(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        object_list.selected_columns = ["value", "other_value"]

        tester = UITester()
        with tester.create_ui(object_list, dict(view=select_columns_view)) \
                as ui:
            values_table = tester.find_by_name(ui, "values")
            selected = values_table.inspect(Selected())

        self.assertEqual(selected, ["value", "other_value"])

    def test_table_editor_select_column_index(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )

        view = select_column_index_view
        tester = UITester()
        with tester.create_ui(object_list, dict(view=view)) as ui:
            # click a cell in the index 1 column
            values_table = tester.find_by_name(ui, "values")
            col1_cell = values_table.locate(Cell(0, 1))
            col1_cell.perform(MouseClick())

            selected = values_table.inspect(SelectedIndices())

        self.assertEqual(selected, [1])

    def test_table_editor_select_column_indices(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        object_list.selected_indices = [0, 1]

        view = select_column_indices_view
        tester = UITester()
        with tester.create_ui(object_list, dict(view=view)) as ui:
            values_table = tester.find_by_name(ui, "values")
            selected = values_table.inspect(SelectedIndices())

        self.assertEqual(selected, [0, 1])

    def test_table_editor_select_cell(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )

        tester = UITester()
        with tester.create_ui(object_list, dict(view=select_cell_view)) as ui:
            # click the cell at (5,0)
            values_table = tester.find_by_name(ui, "values")
            cell_5_0 = values_table.locate(Cell(5, 0))
            cell_5_0.perform(MouseClick())

            selected = values_table.inspect(Selected())

        self.assertEqual(selected, [(object_list.values[5], "value")])
        self.assertIs(selected[0], object_list.selected_cell)

    def test_table_editor_modify_cell_with_tester(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        tester = UITester()
        with tester.create_ui(object_list, dict(view=select_row_view)) as ui:
            wrapper = tester.find_by_name(ui, "values").locate(Cell(5, 0))
            wrapper.perform(MouseClick())             # activate edit mode
            wrapper.perform(KeySequence("abc"))
            self.assertEqual(object_list.selected.value, "abc")

            # second column refers to an Int type
            original = object_list.selected.other_value
            wrapper = tester.find_by_name(ui, "values").locate(Cell(5, 1))
            wrapper.perform(MouseClick())
            wrapper.perform(KeySequence("abc"))       # invalid
            self.assertEqual(object_list.selected.other_value, original)

            for _ in range(3):
                wrapper.perform(KeyClick("Backspace"))
            wrapper.perform(KeySequence("12"))  # now ok
            self.assertEqual(object_list.selected.other_value, 12)

    def test_table_editor_check_display_with_tester(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(other_value=0)]
        )
        tester = UITester()
        with tester.create_ui(object_list, dict(view=select_row_view)) as ui:
            wrapper = tester.find_by_name(ui, "values").locate(Cell(0, 1))

            actual = wrapper.inspect(DisplayedText())
            self.assertEqual(actual, "0")

            object_list.values[0].other_value = 123

            actual = wrapper.inspect(DisplayedText())
            self.assertEqual(actual, "123")

    def test_table_editor_escape_retain_edit(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(other_value=0)]
        )
        tester = UITester()
        with tester.create_ui(object_list, dict(view=select_row_view)) as ui:
            cell = tester.find_by_name(ui, "values").locate(Cell(0, 1))

            cell.perform(MouseClick())
            cell.perform(KeySequence("123"))
            cell.perform(KeyClick("Esc"))  # exit edit mode, did not revert

            self.assertEqual(object_list.values[0].other_value, 123)

    def test_table_editor_select_cells(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        object_list.selected_cells = [
            (object_list.values[5], "value"),
            (object_list.values[6], "other value"),
            (object_list.values[8], "value"),
        ]

        tester = UITester()
        with tester.create_ui(object_list, dict(view=select_cells_view)) as ui:
            values_table = tester.find_by_name(ui, "values")
            selected = values_table.inspect(Selected())

        self.assertEqual(
            selected,
            [
                (object_list.values[5], "value"),
                (object_list.values[6], "other value"),
                (object_list.values[8], "value"),
            ],
        )

    def test_table_editor_select_cell_index(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )

        view = select_cell_index_view
        tester = UITester()
        with tester.create_ui(object_list, dict(view=view)) as ui:
            # click the cell at (5,1)
            values_table = tester.find_by_name(ui, "values")
            cell_5_1 = values_table.locate(Cell(5, 1))
            cell_5_1.perform(MouseClick())
            selected = values_table.inspect(SelectedIndices())

        self.assertEqual(selected, [(5, 1)])

    def test_table_editor_select_cell_indices(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        object_list.selected_cell_indices = [(5, 0), (6, 1), (8, 0)]

        view = select_cell_indices_view
        tester = UITester()
        with tester.create_ui(object_list, dict(view=view)) as ui:
            values_table = tester.find_by_name(ui, "values")
            selected = values_table.inspect(SelectedIndices())

        self.assertEqual(selected, [(5, 0), (6, 1), (8, 0)])

    def test_progress_column(self):
        from traitsui.extras.progress_column import ProgressColumn

        progress_view = View(
            Item(
                "values",
                show_label=False,
                editor=TableEditor(
                    columns=[
                        ObjectColumn(name="value"),
                        ProgressColumn(name="other_value"),
                    ]
                ),
            ),
            buttons=["OK"],
        )
        object_list = ObjectList(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        tester = UITester()
        with tester.create_ui(object_list, dict(view=progress_view)):
            pass

    def test_on_perform_action(self):
        # A test for issue #741, where actions with an on_perform function set
        # would get called twice
        object_list = ObjectList(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        mock_function = Mock()
        action = Action(on_perform=mock_function)

        tester = UITester()
        with tester.create_ui(object_list, dict(view=simple_view)) as ui:
            editor = tester.find_by_name(ui, "values")._target
            editor.set_menu_context(None, None, None)
            editor.perform(action)
        mock_function.assert_called_once()

    def test_edit_on_first_click_false(self):
        object_list = ObjectListWithSelection(
            values=[ListItem(value=str(i ** 2)) for i in range(10)]
        )
        tester = UITester()
        with tester.create_ui(
                object_list,
                dict(view=edit_on_first_click_false_view)
        ) as ui:
            wrapper = tester.find_by_name(ui, "values").locate(Cell(5, 0))
            # single click will not activate edit mode
            wrapper.perform(MouseClick())
            with self.assertRaises(Disabled):
                wrapper.perform(KeySequence("abc"))

            # double click will activate edit mode
            wrapper.perform(MouseDClick())
            wrapper.perform(KeySequence("abc"))
            self.assertEqual(object_list.values[5].value, "abc")



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
Font editor

A Font editor in a Traits UI allows the user to select a font from the
operating system.

Typically, you then pass the Font trait to another UI editor, which uses it to
display text. You can also read the Font trait as a string, or access its
individual attributes (note that these attributes are specific to the UI
toolkit -- QT or WX.)

The default 'simple' Font editor style is usually the most useful and powerful
style - it pops up a font selection dialog which is specific to the OS and
toolkit.

This example also displays some other less common style choices.

Please refer to the `FontEditor API docs`_ for further information.

.. _FontEditor API docs: https://docs.enthought.com/traitsui/api/traitsui.editors.font_editor.html#traitsui.editors.font_editor.FontEditor
"""

from traits.api import HasTraits

from traitsui.api import Item, Group, View, Font


class FontEditorDemo(HasTraits):
    """Defines the main FontEditor demo class."""

    # Define a Font trait to view:
    my_font_trait = Font()

    # Display specification (one Item per editor style):
    font_group = Group(
        Item('my_font_trait', style='simple', label='Simple'),
        Item('_'),
        Item('my_font_trait', style='custom', label='Custom'),
        Item('_'),
        Item('my_font_trait', style='text', label='Text'),
        Item('_'),
        Item('my_font_trait', style='readonly', label='ReadOnly'),
    )

    # Demo view:
    traits_view = View(
        font_group, title='FontEditor', buttons=['OK'], resizable=True
    )


# Create the demo:
demo = FontEditorDemo()


# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" This module provides functions for registering interaction handlers
and location solvers for common Qt GUI components.
"""

from traitsui.testing.tester.command import (
    KeyClick,
    KeySequence,
    MouseClick,
)
from traitsui.testing.tester.query import DisplayedText
from traitsui.testing.tester._ui_tester_registry.qt4 import (
    _interaction_helpers,
)


def register_editable_textbox_handlers(registry, target_class, widget_getter):
    """Register common interactions for an editable textbox (in Qt)

    Parameters
    ----------
    registry : TargetRegistry
        The registry being registered to.
    target_class : subclass of type
        The type of target being wrapped in a UIWrapper on which the
        interaction will be performed.
    widget_getter : callable(wrapper: UIWrapper) -> QWidget
        A callable to return a Qt widget for editing text, i.e. QLineEdit
        or QTextEdit.
    """
    handlers = [
        (
            KeySequence,
            (
                lambda wrapper, interaction: _interaction_helpers.key_sequence_textbox(
                    widget_getter(wrapper), interaction, wrapper.delay
                )
            ),
        ),
        (
            KeyClick,
            (
                lambda wrapper, interaction: _interaction_helpers.key_click_qwidget(
                    widget_getter(wrapper), interaction, wrapper.delay
                )
            ),
        ),
        (
            MouseClick,
            (
                lambda wrapper, _: _interaction_helpers.mouse_click_qwidget(
                    widget_getter(wrapper), wrapper.delay
                )
            ),
        ),
        (
            DisplayedText,
            lambda wrapper, _: _interaction_helpers.displayed_text_qobject(
                widget_getter(wrapper)
            ),
        ),
    ]
    for interaction_class, handler in handlers:
        registry.register_interaction(
            target_class=target_class,
            interaction_class=interaction_class,
            handler=handler,
        )



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from traitsui.testing.tester.command import MouseClick
from traitsui.testing.tester.locator import Index
from traitsui.testing.tester.query import SelectedText
from traitsui.testing.tester._ui_tester_registry._common_ui_targets import (
    BaseSourceWithLocation,
)
from traitsui.testing.tester._ui_tester_registry.qt4._interaction_helpers import (  # noqa
    mouse_click_combobox,
    mouse_click_qwidget,
)
from traitsui.testing.tester._ui_tester_registry._traitsui_ui import (
    register_traitsui_ui_solvers,
)
from traitsui.qt4.instance_editor import CustomEditor, SimpleEditor


def _get_nested_ui_simple(target):
    """Obtains a nested UI within a Simple Instance Editor.

    Parameters
    ----------
    target : instance of SimpleEditor
    """
    return target._dialog_ui


def _get_nested_ui_custom(target):
    """Obtains a nested UI within a Custom Instance Editor.

    Parameters
    ----------
    target : instance of CustomEditor
    """
    return target._ui


def _get_combobox(target):
    """Obtains a nested combobox within an Instance Editor.

    Parameters
    ----------
    target : instance of CustomEditor
    """
    return target._choice


def _click_combobox_index(wrapper, _):
    return mouse_click_combobox(
        combobox=_get_combobox(wrapper._target.source),
        index=wrapper._target.location.index,
        delay=wrapper.delay,
    )


def _get_combobox_text(wrapper, _):
    return _get_combobox(wrapper._target).currentText()


class _IndexedCustomEditor(BaseSourceWithLocation):
    """Wrapper class for CustomEditors with a selection."""

    source_class = CustomEditor
    locator_class = Index
    handlers = [
        (MouseClick, _click_combobox_index),
    ]


def register(registry):
    """Register interactions for the given registry.

    If there are any conflicts, an error will occur.

    Parameters
    ----------
    registry : TargetRegistry
        The registry being registered to.
    """
    _IndexedCustomEditor.register(registry)

    registry.register_interaction(
        target_class=SimpleEditor,
        interaction_class=MouseClick,
        handler=lambda wrapper, _: (
            mouse_click_qwidget(wrapper._target._button, delay=wrapper.delay)
        ),
    )
    register_traitsui_ui_solvers(registry, SimpleEditor, _get_nested_ui_simple)

    registry.register_interaction(
        target_class=CustomEditor,
        interaction_class=SelectedText,
        handler=_get_combobox_text,
    )
    register_traitsui_ui_solvers(registry, CustomEditor, _get_nested_ui_custom)



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the various instance descriptors used by the instance editor and
    instance editor factory classes.
"""

from abc import abstractmethod

from traits.api import (
    ABCHasStrictTraits,
    Str,
    Any,
    Dict,
    Tuple,
    Callable,
    Bool,
)

from .ui_traits import AView

from .helper import user_name_for


class InstanceChoiceItem(ABCHasStrictTraits):

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: User interface name for the item
    name = Str()

    #: View associated with this item
    view = AView

    #: Does this item create new instances?
    is_factory = Bool(False)

    def get_name(self, object=None):
        """Returns the name of the item."""
        return self.name

    def get_view(self):
        """Returns the view associated with the object."""
        return self.view

    @abstractmethod
    def get_object(self):
        """Returns the object associated with the item."""
        pass

    @abstractmethod
    def is_compatible(self, object):
        """Indicates whether a specified object is compatible with the item."""
        pass

    def is_selectable(self):
        """Indicates whether the item can be selected by the user."""
        return True

    def is_droppable(self):
        """Indicates whether the item supports drag and drop."""
        return False


class InstanceChoice(InstanceChoiceItem):

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Object associated with the item
    object = Any()

    #: The name of the object trait containing its user interface name:
    name_trait = Str("name")

    def get_name(self, object=None):
        """Returns the name of the item."""
        if self.name != "":
            return self.name

        name = getattr(self.object, self.name_trait, None)
        if isinstance(name, str):
            return name

        return user_name_for(self.object.__class__.__name__)

    def get_object(self):
        """Returns the object associated with the item."""
        return self.object

    def is_compatible(self, object):
        """Indicates whether a specified object is compatible with the item."""
        return object is self.object


class InstanceFactoryChoice(InstanceChoiceItem):

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Indicates whether an instance compatible with this item can be dragged
    #: and dropped rather than created
    droppable = Bool(False)

    #: Indicates whether the item can be selected by the user
    selectable = Bool(True)

    #: A class (or other callable) that can be used to create an item
    #: compatible with this item
    klass = Callable()

    #: Tuple of arguments to pass to **klass** to create an instance
    args = Tuple()

    #: Dictionary of arguments to pass to **klass** to create an instance
    kw_args = Dict(Str, Any)

    #: Does this item create new instances? This value overrides the default.
    is_factory = True

    def get_name(self, object=None):
        """Returns the name of the item."""
        if self.name != "":
            return self.name

        name = getattr(object, "name", None)
        if isinstance(name, str):
            return name

        if issubclass(type(self.klass), type):
            klass = self.klass
        else:
            klass = self.get_object().__class__

        return user_name_for(klass.__name__)

    def get_object(self):
        """Returns the object associated with the item."""
        return self.klass(*self.args, **self.kw_args)

    def is_droppable(self):
        """Indicates whether the item supports drag and drop."""
        return self.droppable

    def is_compatible(self, object):
        """Indicates whether a specified object is compatible with the item."""
        if issubclass(type(self.klass), type):
            return isinstance(object, self.klass)
        return isinstance(object, self.get_object().__class__)

    def is_selectable(self):
        """Indicates whether the item can be selected by the user."""
        return self.selectable


class InstanceDropChoice(InstanceFactoryChoice):

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Indicates whether an instance compatible with this item can be dragged
    #: and dropped rather than created . This value overrides the default.
    droppable = True

    #: Indicates whether the item can be selected by the user. This value
    #: overrides the default.
    selectable = False

    #: Does this item create new instances? This value overrides the default.
    is_factory = False



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

from traitsui.qt4.table_editor import SimpleEditor

from traitsui.testing.tester.command import (
    MouseClick,
    MouseDClick,
    KeyClick,
    KeySequence,
)
from traitsui.testing.tester.locator import Cell
from traitsui.testing.tester.query import (
    DisplayedText,
    Selected,
    SelectedIndices,
)
from traitsui.testing.tester._ui_tester_registry._common_ui_targets import (
    BaseSourceWithLocation
)
from traitsui.testing.tester._ui_tester_registry.qt4 import (
    _interaction_helpers
)


def _query_table_editor_selected(wrapper, interaction):
    selected = wrapper._target.selected
    if not isinstance(selected, list):
        if selected is None:
            return []
        else:
            return [selected]
    else:
        return selected


def _query_table_editor_selected_indices(wrapper, interaction):
    selected_indices = wrapper._target.selected_indices
    if not isinstance(selected_indices, list):
        if selected_indices == -1:
            return []
        else:
            return [selected_indices]
    else:
        return selected_indices


class _SimpleEditorWithCell(BaseSourceWithLocation):
    source_class = SimpleEditor
    locator_class = Cell
    handlers = [
        (MouseClick, lambda wrapper, _: wrapper._target._mouse_click(
            delay=wrapper.delay)),
        (KeyClick, lambda wrapper, interaction: wrapper._target._key_click(
            key=interaction.key,
            delay=wrapper.delay,)),
        (
            KeySequence,
            lambda wrapper, interaction: wrapper._target._key_sequence(
                sequence=interaction.sequence,
                delay=wrapper.delay,
            )
        ),
        (
            DisplayedText,
            lambda wrapper, _: wrapper._target._get_displayed_text()
        ),
        (MouseDClick, lambda wrapper, _: wrapper._target._mouse_dclick(
            delay=wrapper.delay,)),
    ]

    def _get_model_view_index(self):
        table_view = self.source.table_view
        return dict(
            model=table_view.model(),
            view=table_view,
            index=table_view.model().index(
                self.location.row, self.location.column
            ),
        )

    def _mouse_click(self, delay=0):
        _interaction_helpers.mouse_click_item_view(
            **self._get_model_view_index(),
            delay=delay,
        )

    def _mouse_dclick(self, delay=0):
        _interaction_helpers.mouse_dclick_item_view(
            **self._get_model_view_index(),
            delay=delay,
        )

    def _key_sequence(self, sequence, delay=0):
        _interaction_helpers.key_sequence_item_view(
            **self._get_model_view_index(),
            sequence=sequence,
            delay=delay,
        )

    def _key_click(self, key, delay=0):
        _interaction_helpers.key_click_item_view(
            **self._get_model_view_index(),
            key=key,
            delay=delay,
        )

    def _get_displayed_text(self):
        return _interaction_helpers.get_display_text_item_view(
            **self._get_model_view_index(),
        )


def register(registry):
    """ Register interactions for the given registry.

    If there are any conflicts, an error will occur.

    Parameters
    ----------
    registry : TargetRegistry
        The registry being registered to.
    """
    _SimpleEditorWithCell.register(registry)
    registry.register_interaction(
        target_class=SimpleEditor,
        interaction_class=Selected,
        handler=_query_table_editor_selected
    )
    registry.register_interaction(
        target_class=SimpleEditor,
        interaction_class=SelectedIndices,
        handler=_query_table_editor_selected_indices
    )



"""
This example demonstrates how to test interacting with a TableEditor.

The GUI being tested is written in the demo under the same name (minus the
preceding 'test') in the outer directory.
"""

import os
import runpy
import unittest


from traitsui.testing.api import (
    Cell, KeyClick, KeySequence, MouseClick, UITester
)
from traitsui.tests._tools import requires_toolkit, ToolkitName

#: Filename of the demo script
FILENAME = "TableEditor_demo.py"

#: Path of the demo script
DEMO_PATH = os.path.join(os.path.dirname(__file__), "..", FILENAME)


class TestTableEditorDemo(unittest.TestCase):

    @requires_toolkit([ToolkitName.qt])
    def test_list_editor_demo(self):
        demo = runpy.run_path(DEMO_PATH)["demo"]

        tester = UITester()
        with tester.create_ui(demo) as ui:
            employees_table = tester.find_by_name(ui, "employees")

            # clicking a cell enters edit mode and selects full text
            cell_21 = employees_table.locate(Cell(2, 1))
            cell_21.perform(MouseClick())
            cell_21.perform(KeySequence("Jones"))
            cell_21.perform(KeyClick("Enter"))

            self.assertEqual(demo.employees[0].last_name, 'Jones')

            # third column corresponds to Full Name property
            cell_32 = employees_table.locate(Cell(3, 2))
            cell_32.perform(MouseClick())


# Run the test(s)
unittest.TextTestRunner().run(
    unittest.TestLoader().loadTestsFromTestCase(TestTableEditorDemo)
)



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Demo showing how to use the Windows specific Flash editor.
"""

# Imports:
from traitsui.wx.extra.windows.flash_editor import FlashEditor

from traits.api import Enum, HasTraits

from traitsui.api import View, HGroup, Item

# The demo class:


class FlashDemo(HasTraits):

    # The Flash file to display:
    flash = Enum(
        'http://www.ianag.com/arcade/swf/sudoku.swf',
        'http://www.ianag.com/arcade/swf/f-336.swf',
        'http://www.ianag.com/arcade/swf/f-3D-Reversi-1612.swf',
        'http://www.ianag.com/arcade/swf/game_234.swf',
        'http://www.ianag.com/arcade/swf/flashmanwm.swf',
        'http://www.ianag.com/arcade/swf/2379_gyroball.swf',
        'http://www.ianag.com/arcade/swf/f-1416.swf',
        'http://www.ianag.com/arcade/swf/mah_jongg.swf',
        'http://www.ianag.com/arcade/swf/game_e4fe4e55fedc2f502be627ee6df716c5.swf',
        'http://www.ianag.com/arcade/swf/rhumb.swf',
    )

    # The view to display:
    view = View(
        HGroup(Item('flash', label='Pick a game to play')),
        '_',
        Item('flash', show_label=False, editor=FlashEditor()),
    )


# Create the demo:
demo = FlashDemo()

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Trait definition for a wxPython-based color.
"""

from ast import literal_eval

import wx

from pyface.color import Color as PyfaceColor
from pyface.util.color_helpers import channels_to_ints
from pyface.util.color_parser import color_table
from traits.api import Trait, TraitError


# -------------------------------------------------------------------------
#  W3CColourDatabase
# -------------------------------------------------------------------------


class W3CColourDatabase(object):
    """Proxy for the ColourDatabase which allows for finding W3C colors.

    This class is necessary because the wx 'green' is the W3C 'lime',
    and we need some means to lookup the color names since wx has
    only a few hardcoded.

    This class is a proxy because AddColour expects a wx.ColourDatabase
    instance, not an instance of a subclass
    """

    _database = wx.ColourDatabase()

    def __init__(self):
        # correct for differences in definitions
        self._color_names = [
            "aqua",
            "black",
            "blue",
            "fuchsia",
            "gray",
            "green",
            "lime",
            "maroon",
            "navy",
            "olive",
            "purple",
            "red",
            "silver",
            "teal",
            "white",
            "yellow",
        ]

        self.AddColour("aqua", wx.Colour(0, 0xFF, 0xFF, 255))
        self.AddColour("fuchsia", wx.Colour(0xFF, 0, 0xFF, 255))
        self.AddColour("green", wx.Colour(0, 0x80, 0, 255))
        self.AddColour("lime", wx.Colour(0, 0xFF, 0, 255))
        self.AddColour("maroon", wx.Colour(0x80, 0x0, 0, 255))
        self.AddColour("navy", wx.Colour(0x00, 0x0, 0x80, 255))
        self.AddColour("olive", wx.Colour(0x80, 0x80, 0, 255))
        self.AddColour("purple", wx.Colour(0x80, 0x00, 0x80, 255))
        self.AddColour("silver", wx.Colour(0xC0, 0xC0, 0xC0, 255))
        self.AddColour("teal", wx.Colour(0, 0x80, 0x80, 255))

        # add all the standard colours
        for name, rgba in color_table.items():
            rgba_bytes = channels_to_ints(rgba)
            self.AddColour(name, wx.Colour(*rgba_bytes))

    def AddColour(self, name, color):
        if name not in self._color_names:
            self._color_names.append(name)
        return self._database.AddColour(name, color)

    def Find(self, color_name):
        return self._database.Find(color_name)

    def FindName(self, color):
        for color_name in self._color_names:
            if self.Find(color_name) == color:
                return color_name

        return ""


w3c_color_database = W3CColourDatabase()

# -------------------------------------------------------------------------
#  Convert a number into a wxColour object:
# -------------------------------------------------------------------------


def tuple_to_wxcolor(tup):
    if 3 <= len(tup) <= 4:
        for c in tup:
            if not isinstance(c, int):
                raise TraitError

        return wx.Colour(*tup)
    else:
        raise TraitError


def convert_to_color(object, name, value):
    """Converts a number into a wxColour object."""
    if isinstance(value, tuple):
        return tuple_to_wxcolor(value)

    elif isinstance(value, PyfaceColor):
        return value.to_toolkit()

    elif isinstance(value, wx.Colour):
        return value

    elif isinstance(value, str):
        # Allow for spaces in the string value.
        value = value.replace(" ", "")

        if value in standard_colors:
            return standard_colors[value]

        # Check for tuple string
        try:
            tup = literal_eval(value)
        except Exception:
            raise TraitError
        return tuple_to_wxcolor(tup)

    else:
        try:
            num = int(value)
        except Exception:
            raise TraitError
        return wx.Colour(num // 0x10000, (num // 0x100) & 0xFF, num & 0xFF)

    raise TraitError


convert_to_color.info = (
    "a string of the form (r,g,b) or (r,g,b,a) where r, "
    "g, b, and a are integers from 0 to 255, a wx.Colour "
    "instance, an integer which in hex is of the form "
    "0xRRGGBB, where RR is red, GG is green, and BB is "
    "blue"
)

# -------------------------------------------------------------------------
#  Standard colors:
# -------------------------------------------------------------------------

standard_colors = {}
for name in color_table:
    try:
        wx_color = w3c_color_database.Find(name)
        standard_colors[name] = convert_to_color(None, None, wx_color)
    except:
        pass

# -------------------------------------------------------------------------
#  Define wxPython specific color traits:
# -------------------------------------------------------------------------

### Note: Declare the editor to be a function which returns the ColorEditor
# class from traits ui to avoid circular import issues. For backwards
# compatibility with previous Traits versions, the 'editors' folder in Traits
# project declares 'from api import *' in its __init__.py. The 'api' in turn
# can contain classes that have a Color trait which lead to this file getting
# imported. This leads to a circular import when declaring a Color trait.


def get_color_editor(*args, **traits):
    from .color_editor import ToolkitEditorFactory

    return ToolkitEditorFactory(*args, **traits)


def WxColor(default="white", allow_none=False, **metadata):
    """Defines wxPython-specific color traits."""
    if default is None:
        allow_none = True

    if allow_none:
        return Trait(
            default,
            None,
            standard_colors,
            convert_to_color,
            editor=get_color_editor,
            **metadata,
        )

    return Trait(
        default,
        standard_colors,
        convert_to_color,
        editor=get_color_editor,
        **metadata,
    )



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Defines the tuple editor factory for all traits user interface toolkits.
"""

from traits.api import (
    BaseTuple,
    Bool,
    HasTraits,
    List,
    Str,
    Int,
    Any,
    TraitType,
)
from traits.trait_base import SequenceTypes

from traitsui.editor import Editor
from traitsui.editor_factory import EditorFactory
from traitsui.group import Group
from traitsui.item import Item
from traitsui.view import View


class TupleEditor(EditorFactory):
    """Editor factory for tuple editors."""

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Trait definitions for each tuple field
    types = Any()

    #: Labels for each of the tuple fields
    labels = List(Str)

    #: Editors for each of the tuple fields:
    editors = List(EditorFactory)

    #: Number of tuple fields or rows
    cols = Int(1)

    #: Is user input set on every keystroke? This is applied to every field
    #: of the tuple, provided the field does not already have an 'auto_set'
    #: metadata or an editor defined.
    auto_set = Bool(True)

    #: Is user input set when the Enter key is pressed? This is applied to
    #: every field of the tuple, provided the field does not already have an
    #: 'enter_set' metadata or an editor defined.
    enter_set = Bool(False)


class SimpleEditor(Editor):
    """Simple style of editor for tuples.

    The editor displays an editor for each of the fields in the tuple, based on
    the type of each field.
    """

    def init(self, parent):
        """Finishes initializing the editor by creating the underlying toolkit
        widget.
        """
        self._ts = ts = TupleStructure(self)
        self._ui = ui = ts.view.ui(ts, parent, kind="subpanel").trait_set(
            parent=self.ui
        )
        self.control = ui.control
        self.set_tooltip()

    def update_editor(self):
        """Updates the editor when the object trait changes external to the
        editor.
        """
        ts = self._ts
        for i, value in enumerate(self.value):
            setattr(ts, "f%d" % i, value)

    def get_error_control(self):
        """Returns the editor's control for indicating error status."""
        return self._ui.get_error_controls()


class TupleStructure(HasTraits):
    """Creates a view containing items for each field in a tuple."""

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Editor this structure is linked to
    editor = Any()

    #: The constructed View for the tuple
    view = Any()

    #: Number of tuple fields
    fields = Int()

    def __init__(self, editor):
        """Initializes the object."""
        super().__init__(editor=editor)

        factory = editor.factory
        types = factory.types
        labels = factory.labels
        editors = factory.editors
        cols = factory.cols

        # Get the tuple we are mirroring:
        object = editor.value

        # For each tuple field, add a trait with the appropriate trait
        # definition and default value:
        content = []
        self.fields = len(object)
        len_labels = len(labels)
        len_editors = len(editors)

        if types is None:
            type = editor.value_trait.handler
            if isinstance(type, BaseTuple):
                types = type.types

        if not isinstance(types, SequenceTypes):
            types = [types]

        len_types = len(types)
        if len_types == 0:
            types = [Any]
            len_types = 1

        for i, value in enumerate(object):
            type = types[i % len_types]

            auto_set = enter_set = None
            if isinstance(type, TraitType):
                auto_set = type.auto_set
                enter_set = type.enter_set
            if auto_set is None:
                auto_set = editor.factory.auto_set
            if enter_set is None:
                enter_set = editor.factory.enter_set

            label = ""
            if i < len_labels:
                label = labels[i]

            field_editor = None
            if i < len_editors:
                field_editor = editors[i]

            name = "f%d" % i
            self.add_trait(
                name,
                type(
                    value,
                    event="field",
                    auto_set=auto_set,
                    enter_set=enter_set,
                ),
            )
            item = Item(name=name, label=label, editor=field_editor)
            if cols <= 1:
                content.append(item)
            else:
                if (i % cols) == 0:
                    group = Group(orientation="horizontal")
                    content.append(group)

                group.content.append(item)

        self.view = View(Group(show_labels=(len_labels != 0), *content))

    def _field_changed(self, name, old, new):
        """Updates the underlying tuple when any field changes value."""
        index = int(name[1:])
        value = self.editor.value
        if new != value[index]:
            self.editor.value = tuple(
                [getattr(self, "f%d" % i) for i in range(self.fields)]
            )


# This alias is deprecated and will be removed in TraitsUI 8.
ToolkitEditorFactory = TupleEditor



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Traits UI 'display only' image editor.
"""


from pyface.image_resource import ImageResource

from traitsui.ui_traits import convert_bitmap

# FIXME: ImageEditor is a proxy class defined here just for backward
# compatibility. The class has been moved to the
# traitsui.editors.image_editor file.
from traitsui.editors.image_editor import ImageEditor

from .editor import Editor

from .image_control import ImageControl

# -------------------------------------------------------------------------
#  '_ImageEditor' class:
# -------------------------------------------------------------------------


class _ImageEditor(Editor):
    """Traits UI 'display only' image editor."""

    def init(self, parent):
        """Finishes initializing the editor by creating the underlying toolkit
        widget.
        """
        image = self.factory.image
        if image is None:
            image = self.value
        if image is not None:
            bitmap = convert_bitmap(image)
        else:
            bitmap = None

        self.control = ImageControl(parent, bitmap, padding=0)

        self.set_tooltip()

    def update_editor(self):
        """Updates the editor when the object trait changes externally to the
        editor.
        """
        if self.factory.image is None:
            value = self.value
            if value is not None:
                bitmap = convert_bitmap(value)
            else:
                bitmap = None
            self.control.Bitmap(bitmap)



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Demo showing how to use the Windows specific Internet Explorer editor.
"""

# Imports:
from traitsui.wx.extra.windows.ie_html_editor import IEHTMLEditor

from traits.api import Str, List, Button, HasTraits

from traitsui.api import View, VGroup, HGroup, Item, TextEditor, ListEditor

# The web page class:


class WebPage(HasTraits):

    # The URL to display:
    url = Str('http://code.enthought.com')

    # The page title:
    title = Str()

    # The page status:
    status = Str()

    # The browser navigation buttons:
    back = Button('<--')
    forward = Button('-->')
    home = Button('Home')
    stop = Button('Stop')
    refresh = Button('Refresh')
    search = Button('Search')

    # The view to display:
    view = View(
        HGroup(
            'back',
            'forward',
            'home',
            'stop',
            'refresh',
            'search',
            '_',
            Item('status', style='readonly'),
            show_labels=False,
        ),
        Item(
            'url',
            show_label=False,
            editor=IEHTMLEditor(
                home='home',
                back='back',
                forward='forward',
                stop='stop',
                refresh='refresh',
                search='search',
                title='title',
                status='status',
            ),
        ),
    )


# The demo class:


class InternetExplorerDemo(HasTraits):

    # A URL to display:
    url = Str('http://')

    # The list of web pages being browsed:
    pages = List(WebPage)

    # The view to display:
    view = View(
        VGroup(
            Item(
                'url',
                label='Location',
                editor=TextEditor(auto_set=False, enter_set=True),
            )
        ),
        Item(
            'pages',
            show_label=False,
            style='custom',
            editor=ListEditor(
                use_notebook=True,
                deletable=True,
                dock_style='tab',
                export='DockWindowShell',
                page_name='.title',
            ),
        ),
    )

    # Event handlers:
    def _url_changed(self, url):
        self.pages.append(WebPage(url=url.strip()))


# Create the demo:
demo = InternetExplorerDemo(
    pages=[
        WebPage(url='http://code.enthought.com/projects/traits/'),
        WebPage(url='http://dmorrill.com'),
    ]
)

# Run the demo (if invoked from the command line):
if __name__ == '__main__':
    demo.configure_traits()



# (C) Copyright 2004-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

"""
A simple demonstration of how to use the ImageEditor to add a graphic element
to a Traits UI View.

This example needs NumPy to show an example of a dynamically created image.
"""

# Imports:
from os.path import join, dirname

import numpy as np

from pyface.api import ArrayImage, Image, ImageResource
from traits.api import HasTraits, Str
from traitsui.api import View, VGroup, Item, ImageEditor

# Constants:

# The images folder is in the same folder as this file:
search_path = [dirname(__file__)]


# Define the demo class:
class Employee(HasTraits):

    # Define the traits:
    name = Str()
    dept = Str()
    email = Str()
    picture = Image()
    gradient = Image()

    # Define the view:
    view = View(
        VGroup(
            VGroup(
                Item(
                    'name',
                    show_label=False,
                    editor=ImageEditor(
                        image=ImageResource('info', search_path=search_path)
                    ),
                )
            ),
            VGroup(
                Item('name'),
                Item('dept'),
                Item('email'),
                Item(
                    'picture',
                    editor=ImageEditor(
                        scale=True,
                        preserve_aspect_ratio=True,
                        allow_upscaling=True,
                    ),
                    springy=True,
                ),
                Item(
                    'gradient',
                    editor=ImageEditor(
                        scale=True,
                        preserve_aspect_ratio=True,
                        allow_upscaling=True,
                    ),
                    springy=True,
                ),
            ),
        ),
        resizable=True,
    )


# generate a 2D NumPy array of RGB values
gradient = np.empty(shape=(256, 256, 3), dtype='uint8')
gradient[:, :, 0] = np.arange(256).reshape(256, 1)
gradient[:, :, 1] = np.arange(256).reshape(1, 256)
gradient[:, :, 2] = np.arange(255, -1, -1).reshape(1, 256)

# Create the demo:
popup = Employee(
    name='William Murchison',
    dept='Receiving',
    email='wmurchison@acme.com',
    picture=ImageResource('python-logo', search_path=search_path),
    gradient=ArrayImage(data=gradient),
)

# Run the demo (if invoked form the command line):
if __name__ == '__main__':
    popup.configure_traits()



# (C) Copyright 2008-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

# ------------------------------------------------------------------------------
# Copyright (c) 2007, Riverbank Computing Limited
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD license.
# However, when used with the GPL version of PyQt the additional terms
# described in the PyQt GPL exception also apply

#
# Author: Riverbank Computing Limited
# ------------------------------------------------------------------------------

""" Defines the key binding editor for use with the KeyBinding class. This is a
specialized editor used to associate a particular key with a control (i.e., the
key binding editor).
"""


from pyface.qt import QtCore, QtGui

from pyface.api import YES, confirm
from traits.api import Bool, Event

from .editor import Editor

from .key_event_to_name import key_event_to_name


class KeyBindingEditor(Editor):
    """An editor for modifying bindings of keys to controls."""

    # -------------------------------------------------------------------------
    #  Trait definitions:
    # -------------------------------------------------------------------------

    #: Does the editor's control have focus currently?
    has_focus = Bool(False)

    #: Keyboard event
    key = Event()

    #: Clear field event
    clear = Event()

    def init(self, parent):
        """Finishes initializing the editor by creating the underlying toolkit
        widget.
        """
        self.control = KeyBindingCtrl(self)

    def update_editor(self):
        """Updates the editor when the object trait changes externally to the
        editor.
        """
        self.control.setText(self.value)

    def _key_changed(self, event):
        """Handles a keyboard event."""
        binding = self.object
        key_name = key_event_to_name(event)
        cur_binding = self.ui.parent.handler.key_binding_for(binding, key_name)
        if cur_binding is not None:
            result = confirm(
                parent=self.control,
                message=(
                    f"{key_name!r} has already been assigned to "
                    f"'{cur_binding.description}'.\n"
                    "Do you wish to continue?"
                ),
                title="Duplicate Key Definition",
            )
            if result != YES:
                return

        self.value = key_name
        # Need to manually update editor because the update to the value
        # won't get transferred to toolkit object due to loopback protection.
        self.update_editor()

    def _clear_changed(self):
        """Handles a clear field event."""
        self.value = ""
        # Need to manually update editor because the update to the value
        # won't get transferred to toolkit object due to loopback protection.
        self.update_editor()


class KeyBindingCtrl(QtGui.QLineEdit):
    """PyQt control for editing key bindings."""

    def __init__(self, editor, parent=None):
        super().__init__(parent)

        self.setFocusPolicy(QtCore.Qt.FocusPolicy.StrongFocus)
        self.setMinimumWidth(160)
        self.setMaximumWidth(160)
        self.setReadOnly(True)

        # Save the reference to the controlling editor object:
        self.editor = editor

        # Indicate we don't have the focus right now:
        editor.has_focus = False

    def keyPressEvent(self, event):
        """Handle keyboard keys being pressed."""
        # Ignore presses of the control and shift keys.
        if event.key() not in (QtCore.Qt.Key.Key_Control, QtCore.Qt.Key.Key_Shift):
            self.editor.key = event

    def paintEvent(self, event):
        """Updates the screen."""
        super().paintEvent(event)

        if self.editor.has_focus:
            w = self.width()
            h = self.height()
            p = QtGui.QPainter(self)

            p.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing, True)
            pen = QtGui.QPen(QtGui.QColor("tomato"))
            pen.setWidth(2)
            p.setPen(pen)
            p.drawRect(1, 1, w - 2, h - 2)

            p.end()

    def focusInEvent(self, event):
        """Handles getting the focus."""
        self.editor.has_focus = True
        self.update()

    def focusOutEvent(self, event):
        """Handles losing the focus."""
        self.editor.has_focus = False
        self.update()

    def mouseDoubleClickEvent(self, event):
        """Handles the user double clicking the control to clear its contents."""
        self.editor.clear = True



# (C) Copyright 2008-2023 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest

from pyface.font import Font as PyfaceFont
from pyface.util.font_parser import simple_parser
from pyface.qt.QtGui import QFont, QFontInfo, QApplication
from traits.api import HasTraits, TraitError

from ..font_trait import (
    PyQtFont, TraitsFont, create_traitsfont, font_families, font_styles,
    font_to_str, font_weights
)


class FontExample(HasTraits):

    font = PyQtFont()


class TestPyQtFont(unittest.TestCase):

    def test_create_traitsfont(self):
        expected_outcomes = {}
        expected_outcomes[""] = TraitsFont()

        for weight, qt_weight in font_weights.items():
            expected_outcomes[weight] = TraitsFont()
            expected_outcomes[weight].setWeight(qt_weight)

        for style, qt_style in font_styles.items():
            expected_outcomes[style] = TraitsFont()
            expected_outcomes[style].setStyle(qt_style)

        expected_outcomes["underline"] = TraitsFont()
        expected_outcomes["underline"].setUnderline(True)

        expected_outcomes["18"] = TraitsFont()
        expected_outcomes["18"].setPointSize(18)
        expected_outcomes["18 pt"] = TraitsFont()
        expected_outcomes["18 pt"].setPointSize(18)
        expected_outcomes["18 point"] = TraitsFont()
        expected_outcomes["18 point"].setPointSize(18)

        for family, qt_style_hint in font_families.items():
            expected_outcomes[family] = TraitsFont()
            expected_outcomes[family].setStyleHint(qt_style_hint)

        default_size = QApplication.font().pointSize()
        expected_outcomes["Courier"] = TraitsFont("Courier", default_size)
        expected_outcomes["Comic Sans"] = TraitsFont("Comic Sans", default_size)
        expected_outcomes["18 pt Bold Oblique Underline Comic Sans script"] = TraitsFont(
            "Comic Sans", 18, QFont.Weight.Bold, False
        )
        expected_outcomes["18 pt Bold Oblique Underline Comic Sans script"].setStyleHint(QFont.StyleHint.Cursive)
        expected_outcomes["18 pt Bold Oblique Underline Comic Sans script"].setStyle(QFont.Style.StyleOblique)
        expected_outcomes["18 pt Bold Oblique Underline Comic Sans script"].setUnderline(True)

        for name, expected in expected_outcomes.items():
            with self.subTest(name=name):
                result = create_traitsfont(name)

                # test we get expected font
                self.assertIsInstance(result, TraitsFont)
                self.assert_qfont_equal(result, expected)

                # round-trip through font_to_str
                result_2 = create_traitsfont(font_to_str(result))
                self.assert_qfont_equal(result, result_2)

    def test_create_traitsfont_qfont(self):
        font = QFont("Comic Sans", 18, QFont.Weight.Bold, False)
        traits_font = create_traitsfont(font)

        self.assertIsInstance(traits_font, TraitsFont)
        self.assert_qfont_equal(traits_font, font)

    def test_create_traitsfont_pyface_font(self):
        args = simple_parser("18 pt Bold Oblique Underline Courier")
        font = PyfaceFont(**args)
        traits_font = create_traitsfont(font)

        self.assertIsInstance(traits_font, TraitsFont)
        self.assert_qfont_equal(traits_font, font.to_toolkit())

    def test_font_trait_default(self):
        obj = FontExample()

        self.assertIsInstance(obj.font, TraitsFont)
        self.assert_qfont_equal(obj.font, TraitsFont())

    def test_font_trait_str(self):
        obj = FontExample(font="18 pt Bold Oblique Underline Comic Sans script")
        qfont = TraitsFont(
            "Comic Sans", 18, QFont.Weight.Bold, False
        )
        qfont.setStyleHint(QFont.StyleHint.Cursive)
        qfont.setStyle(QFont.Style.StyleOblique)
        qfont.setUnderline(True)

        self.assertIsInstance(obj.font, TraitsFont)
        self.assert_qfont_equal(obj.font, qfont)

    def test_font_trait_qfont(self):
        qfont = TraitsFont(
            "Comic Sans", 18, QFont.Weight.Bold, False
        )
        qfont.setStyleHint(QFont.StyleHint.Cursive)
        qfont.setStyle(QFont.Style.StyleOblique)
        qfont.setUnderline(True)
        obj = FontExample(font=qfont)

        self.assertIsInstance(obj.font, TraitsFont)
        self.assert_qfont_equal(obj.font, qfont)

    def test_font_trait_pyface_font(self):
        args = simple_parser("18 pt Bold Oblique Underline Courier typewriter")
        font = PyfaceFont(**args)
        obj = FontExample(font=font)

        self.assertIsInstance(obj.font, TraitsFont)
        self.assert_qfont_equal(obj.font, font.to_toolkit())

    def test_font_trait_none(self):
        obj = FontExample(font=None)

        self.assertIsNone(obj.font)

    def test_font_trait_bad(self):
        with self.assertRaises(TraitError):
            obj = FontExample(font=1)

    def test_traits_font_reduce(self):
        traits_font = TraitsFont(
            "Comic Sans", 18, QFont.Weight.Bold, False
        )
        traits_font.setStyleHint(QFont.StyleHint.Cursive)
        traits_font.setStyle(QFont.Style.StyleOblique)
        traits_font.setUnderline(True)

        result = traits_font.__reduce_ex__(None)

        self.assertEqual(
            result,
            (
                create_traitsfont,
                ("18 point Comic Sans Oblique Bold underline",),
            ),
        )

    def test_traits_font_str(self):
        traits_font = TraitsFont(
            "Comic Sans", 18, QFont.Weight.Bold, False
        )
        traits_font.setStyleHint(QFont.StyleHint.Cursive)
        traits_font.setStyle(QFont.Style.StyleOblique)
        traits_font.setUnderline(True)

        result = str(traits_font)

        self.assertEqual(
            result,
            "18 point Comic Sans Oblique Bold underline",
        )

    def assert_qfont_equal(self, font, other):
        self.assertIsInstance(font, QFont)
        self.assertIsInstance(other, QFont)

        self.assertEqual(font.family(), other.family())
        self.assertEqual(font.styleHint(), font.styleHint())
        self.assertEqual(font.pointSize(), other.pointSize())
        self.assertEqual(font.style(), other.style())
        self.assertEqual(font.weight(), other.weight())
        self.assertEqual(font.underline(), other.underline())
